/* 
 * enPathia
 * Eneso Tecnología de Adaptación, S.L. 
 */

#include <cairo.h>
#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include <glib/gstdio.h>

#ifdef G_OS_WIN32
#define _WIN32_WINNT 0x0501
#include <windows.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "edecor.h"
#include "action.h"
#include "dictionary.h"
#include "enpathia.h"
#include "keys.h"
#include "misc.h"
#include "profile.h"
#include "protocol.h"
#include "window.h"
#include "keyboard.h"

#define KEYB_ROWS		4
#define DICT_ROWS		3
#define KEYB_COLS		10
#define DICT_COLS		4

#define KEY_COLOR		0.1
#define DICT_COLOR		0.0

#define N_SPECIAL_KEYS	10

#ifndef __MACH__
#define BAR_HEIGHT		22
#else
#define BAR_HEIGHT		0
#endif	/* __MACH__ */

#define WRITTEN_TEXT_BUFFER_SIZE	100

static const gboolean layout_has_uppercase [N_KEYB_LAYOUTS] =
{ 
	/* 	en_US 	en_UK 	el		es 		he		de		it		pt */
	TRUE, 	TRUE, 	TRUE,	TRUE, 	FALSE, 	TRUE, 	TRUE, 	TRUE
};
static const gboolean layout_has_symbols_2 [N_KEYB_LAYOUTS] = 
{ 
	/* 	en_US 	en_UK 	el		es 		he		de		it		pt */
	FALSE, 	FALSE, 	FALSE,	TRUE, 	FALSE, 	TRUE, 	FALSE, 	FALSE
};

static const guint	keyb_width [N_KEYB_SIZES]				= { 400, 600, 800 };
static const guint	keyb_height [N_KEYB_SIZES]				= {
	258 + BAR_HEIGHT, 387 + BAR_HEIGHT, 516 + BAR_HEIGHT };
static const guint	keyb_top_height [N_KEYB_SIZES]			= {
	168 + BAR_HEIGHT, 252 + BAR_HEIGHT, 336 + BAR_HEIGHT };
static const guint	keyb_bottom_height [N_KEYB_SIZES]		= { 90, 135, 180 };

static const guint	x_keys [N_KEYB_SIZES]					= { 11, 17, 22 };
static const guint	y_keys [N_KEYB_SIZES]					= {
	10 + BAR_HEIGHT, 15 + BAR_HEIGHT, 20 + BAR_HEIGHT };
static const guint	key_width [N_KEYB_SIZES]				= { 34, 51, 68 };
static const guint	key_height [N_KEYB_SIZES]				= { 32, 48, 64 };
static const guint	keys_x_spacing [N_KEYB_SIZES]			= { 38, 57, 76 };
static const guint	keys_y_spacing [N_KEYB_SIZES]			= { 36, 54, 72 };
static const guint	radius_key_border [N_KEYB_SIZES]		= { 8, 12, 16 };

static const guint	x_dict_item [N_KEYB_SIZES]				= { 11, 17, 22 };
static const guint	y_dict_item [N_KEYB_SIZES]				= {
	174 + BAR_HEIGHT, 261 + BAR_HEIGHT, 348 + BAR_HEIGHT };
static const guint	dict_width [N_KEYB_SIZES]				= { 91, 136.5, 182 };
static const guint	dict_height [N_KEYB_SIZES]				= { 22, 33, 44 };
static const gfloat	dict_x_spacing [N_KEYB_SIZES]			= { 94, 143, 193 };
static const guint	dict_y_spacing [N_KEYB_SIZES]			= { 24, 36, 48 };
static const guint	dict_x_text_offset [N_KEYB_SIZES]		= { 6, 9, 12 };
static const guint	dict_y_text_offset [N_KEYB_SIZES]		= { 8, 12, 16 };
static const guint	radius_dict_border [N_KEYB_SIZES]		= { 2, 3, 4};

static const guint	layout_bar_border [N_KEYB_SIZES]		= { 5, 8, 10 };
static const guint	layout_bar_offset [N_KEYB_SIZES]		= { 5, 8, 10 };
static const guint	layout_key_width [N_KEYB_SIZES]			= { 40, 60, 80 };
static const guint	layout_key_height [N_KEYB_SIZES]		= { 18, 27, 36 };
static const gfloat	layout_key_y_spacing [N_KEYB_SIZES]		= { 18.75, 30, 40 };
static const guint	radius_layout_bar_border [N_KEYB_SIZES]	= { 4, 6, 8 };

static const guint	x_led_ctrl [N_KEYB_SIZES]				= { 20, 30, 40 };
static const guint	x_led_alt [N_KEYB_SIZES]				= { 60, 90, 120 };
static const guint	x_led_shift [N_KEYB_SIZES]				= { 100, 150, 200 };
static const guint	y_leds [N_KEYB_SIZES]					= {
	158 + BAR_HEIGHT, 237 + BAR_HEIGHT, 316 + BAR_HEIGHT };
static const guint	radius_led [N_KEYB_SIZES]				= { 4, 6, 8 };

static const gchar*	font_key [N_KEYB_SIZES]				= { "Sans 15px", "Sans 23px", "Sans 29px" };
static const gchar*	font_system [N_KEYB_SIZES]			= { "Sans 12px", "Sans 18px", "Sans 24px" };
static const gchar*	font_dict [N_KEYB_SIZES]			= { "Sans 11px", "Sans 16px", "Sans 22px" };
static const gchar*	font_layout_key [N_KEYB_SIZES]		= { "Sans 11px", "Sans 16px", "Sans 22px" };
static const gchar*	font_layout_key_act [N_KEYB_SIZES]	= { "Sans Bold 11px", "Sans Bold 16px", "Sans Bold 22px" };
static const gchar*	font_led [N_KEYB_SIZES]				= { "Sans 10px", "Sans 15px", "Sans 21px" };

static gint size = 0;
static gint loaded_keyb_layout;

enum special_key_index
{
	sk_letters = 0,
	sk_numbers,
	sk_symbols,
	sk_system,
	sk_backspace,
	sk_return,
	sk_close,
	sk_uppercase,
	sk_more_symbols
};

gboolean showing_keyboard		= FALSE;
gboolean mouse_over_keyboard	= FALSE;


/* Images */
static struct
{
	cairo_surface_t*	backspace;
	cairo_surface_t*	close;
	cairo_surface_t*	return_key;
	cairo_surface_t*	letters;
	cairo_surface_t*	uppercase;
	cairo_surface_t*	lowercase;
	cairo_surface_t*	numbers;
	cairo_surface_t*	symbols;
	cairo_surface_t*	system;
	cairo_surface_t*	os_layout;
	cairo_surface_t*	more_symbols;
	cairo_surface_t*	operating_system;
	cairo_surface_t*	options;
	cairo_surface_t*	left;
	cairo_surface_t*	right;
	cairo_surface_t*	up;
	cairo_surface_t*	down;
}
images;

/* Layout */
typedef struct _Key	Key;
struct _Key
{
	gchar*	id;

	gchar*	text;
	guchar	code;
	guchar	modifiers;

	struct _Key*	dead_key;
	gchar*			dead_key_id;

	cairo_surface_t* img;
};

static GSequence*	keyboard_layout = NULL;

/* Keyboards */
typedef struct
{
	gint 	columns;
	gint 	offset_cols;

	Key*	key [KEYB_ROWS * KEYB_COLS];

	gfloat	x_text [KEYB_ROWS * KEYB_COLS];
	gfloat 	y_text [KEYB_ROWS * KEYB_COLS];
	gfloat 	x_key [KEYB_ROWS * KEYB_COLS];
	gfloat	y_key [KEYB_ROWS * KEYB_COLS];
}
Keyboard;

static Keyboard keyboard_lowercase;
static Keyboard keyboard_uppercase;
static Keyboard keyboard_numbers;
static Keyboard keyboard_symbols_1;
static Keyboard keyboard_symbols_2;
static Keyboard keyboard_system;

static Keyboard* current_keyboard;

/* Modifiers */
static gchar modifiers;

/* Dictionary */
static GString*	words;
static gchar*	dictionary [DICT_ROWS * DICT_COLS];

static gint	x_dict [DICT_ROWS * DICT_COLS];
static gint	y_dict [DICT_ROWS * DICT_COLS];

static gint	add_to_dict_entry	= -1;

/* Language keys */
static gint	x_layout_key [N_KEYB_LAYOUTS + 1];
static gint	y_layout_key [N_KEYB_LAYOUTS + 1];
static gint layout_bar_width;

/* Special keys */
typedef struct
{
	gint		row, col;
	gboolean	visible;

	cairo_surface_t*	img;
}
SpecialKey;

SpecialKey special_keys [N_SPECIAL_KEYS];


static gint selected_row, selected_col;
static gboolean active;

static gchar	written_text [WRITTEN_TEXT_BUFFER_SIZE];
static gint		written_text_n_chars;

static gboolean first_time;

static GtkWidget* window;

static gboolean	mouse_over_bar	= FALSE;
static gboolean	dragging_kb		= FALSE;

static gboolean 
window_expose_event_cb		(GtkWidget *widget, GdkEventExpose *event, gpointer data);
static gboolean 
window_motion_event_cb		(GtkWidget *widget, GdkEventMotion *event, gpointer data);
static gboolean
window_button_cb			(GtkWidget *widget, GdkEventButton *event, gpointer data);
static gboolean 
window_enter_event_cb		(GtkWidget *widget, GdkEventCrossing *event, gpointer data);
static gboolean 
window_leave_event_cb		(GtkWidget *widget, GdkEventCrossing *event, gpointer data);
static gboolean
window_configure_event_cb	(GtkWidget *widget, GdkEventConfigure *event, gpointer data);

static void
pointer_at_position	(gint x, gint y);

static void
move_keyboard	(gint x, gint y);

static void
type_key		(Key* key);

static void
draw_key		(cairo_t* cr, gfloat x, gfloat y, gfloat width, gboolean selected);
static void
draw_dict_entry	(cairo_t* cr, gfloat x, gfloat y, gboolean selected);
static void
draw_layout_key	(cairo_t* cr, gfloat x, gfloat y, gboolean current, gboolean selected);
static void
draw_led		(cairo_t* cr, gchar* text, gfloat x, gfloat y, gboolean on);

static void
load_keyboards				(void);
static gint
load_layout					(GSequence** layout, gchar* file);
static gint
load_keyboard				(Keyboard* keyboard, GSequence* layout, gchar* file);
static void
free_keyboards				(void);
static void
calculate_text_positions	(Keyboard* keyboard);
static void
free_key					(gpointer data);
static Key*
find_key					(gchar* id, GSequence* layout);

static void
init_dictionary_entries	(void);
static void
load_dictionaries		(void);

static void
init_special_keys		(void);
static void
change_current_keyboard	(Keyboard* keyboard);
static void
execute_selected_key	(void);


	void
create_keyboard (void)
{
	gchar*	image_file;

	/* In Linux the keyboard has to be toplevel, otherwise it cannot be moved */
#ifdef G_OS_WIN32
	window = gtk_window_new (GTK_WINDOW_POPUP);
#else
	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
#endif

	gtk_window_set_title (GTK_WINDOW (window), "enPathia");
	gtk_window_set_decorated (GTK_WINDOW (window), FALSE);
	gtk_widget_set_app_paintable (window, TRUE);
	gtk_window_set_default_size (GTK_WINDOW (window), keyb_width [size], keyb_height [size]);
	gtk_window_set_keep_above (GTK_WINDOW (window), TRUE);

	/* MacOSX does not send motion notifications of unactive windows, so
	 * we do not highlight the key under the pointer. */

	gtk_widget_add_events (window,
			GDK_STRUCTURE_MASK		|
#ifndef __MACH__
			GDK_POINTER_MOTION_MASK	|
#endif	/* __MACH__ */
			GDK_BUTTON_PRESS_MASK	|
			GDK_ENTER_NOTIFY_MASK	|
			GDK_LEAVE_NOTIFY_MASK);	

	g_signal_connect (window,
			"expose-event", G_CALLBACK (window_expose_event_cb), NULL);
	g_signal_connect (window,
			"configure-event", G_CALLBACK (window_configure_event_cb), NULL);		
#ifndef __MACH__
	g_signal_connect (window,
			"motion-notify-event", G_CALLBACK (window_motion_event_cb), NULL);		
#endif	/* __MACH__ */
	g_signal_connect (window, 
			"button-press-event", G_CALLBACK (window_button_cb), NULL);
	g_signal_connect (window, 
			"enter-notify-event", G_CALLBACK (window_enter_event_cb), NULL);		
	g_signal_connect (window,
			"leave-notify-event", G_CALLBACK (window_leave_event_cb), NULL);		

#ifndef G_OS_WIN32
	GdkScreen*		screen;
	GdkColormap*	colormap;

	/* Set RGBA colormap */ 
	screen = gtk_widget_get_screen (window);
	colormap = gdk_screen_get_rgba_colormap (screen);

	gtk_widget_set_colormap (window, colormap);
#endif

	/* Load the images */
#define LOAD_IMAGE(name,file) \
	image_file = get_data_file ("/images/keyboard/" file);	\
	images.name = cairo_image_surface_create_from_png (image_file);	\
	g_free (image_file);

	LOAD_IMAGE (backspace, "backspace.png");
	LOAD_IMAGE (close, "close.png");
	LOAD_IMAGE (return_key, "return_key.png");
	LOAD_IMAGE (letters, "letters.png");
	LOAD_IMAGE (uppercase, "uppercase.png");
	LOAD_IMAGE (lowercase, "lowercase.png");
	LOAD_IMAGE (numbers, "numbers.png");
	LOAD_IMAGE (symbols, "symbols.png");
	LOAD_IMAGE (system, "system.png");
	LOAD_IMAGE (os_layout, "os_layout.png");
	LOAD_IMAGE (more_symbols, "more_symbols.png");
	LOAD_IMAGE (operating_system, "operating_system.png");
	LOAD_IMAGE (options, "options.png");
	LOAD_IMAGE (left, "left.png");
	LOAD_IMAGE (right, "right.png");
	LOAD_IMAGE (up, "up.png");
	LOAD_IMAGE (down, "down.png");
#undef LOAD_IMAGE

	reset_keyboard ();
}

	void
reset_keyboard (void)
{
	/* Load the dictionaries */
	load_dictionaries ();

	/* Initialize the rest of the keyboard */
	if (keyboard_layout)
	{
		g_sequence_free (keyboard_layout);
		keyboard_layout = NULL;
	}

	load_keyboards ();
	init_dictionary_entries ();
	init_special_keys ();

	change_current_keyboard (&keyboard_lowercase);
	set_keyboard_size (size);
}

	void
destroy_keyboard (void)
{
	/* Free the images */
#define FREE_IMAGE(name) \
	cairo_surface_destroy (images.name);

	FREE_IMAGE (backspace);
	FREE_IMAGE (close);
	FREE_IMAGE (return_key);
	FREE_IMAGE (letters);
	FREE_IMAGE (uppercase);
	FREE_IMAGE (lowercase);
	FREE_IMAGE (numbers);
	FREE_IMAGE (symbols);
	FREE_IMAGE (system);
	FREE_IMAGE (os_layout);
	FREE_IMAGE (more_symbols);
	FREE_IMAGE (operating_system);
	FREE_IMAGE (options);
	FREE_IMAGE (left);
	FREE_IMAGE (right);
	FREE_IMAGE (up);
	FREE_IMAGE (down);
#undef FREE_IMAGE

	g_sequence_free (keyboard_layout);
}

	void
show_keyboard (gint x, gint y)
{
	gint i;

	showing_keyboard = TRUE;

	/* Initial keyboard */
	change_current_keyboard (&keyboard_lowercase);

	/* Delete written text and dictionary */
	init_dictionary_entries ();

	active = FALSE;
	modifiers = 0;
	first_time = TRUE;

	/* Move the window */
	move_keyboard (x, y);

	/* Show the window */
	gtk_window_present (GTK_WINDOW (window));

#ifdef __MACH__
	gtk_window_set_keep_above (GTK_WINDOW (window), TRUE);		
#endif
}

	void
hide_keyboard (void)
{
	showing_keyboard = FALSE;
	gtk_widget_hide (window);
}

	void
set_keyboard_size (gint keyboard_size)
{
	gint	i, j;
	gint	x, y;

	if (keyboard_size < 0)
		keyboard_size = 0;
	else if (keyboard_size > N_KEYB_SIZES - 1)
		keyboard_size = N_KEYB_SIZES - 1;

	size = keyboard_size;

	if (current_profile.config.keyb_show_layout_bar)
		layout_bar_width = layout_key_width [size] + (layout_bar_border [size] << 1);
	else
		layout_bar_width = 0;

	/* Resize and move the window */
	gtk_window_resize (GTK_WINDOW (window), keyb_width [size] + layout_bar_width,
			current_profile.config.keyb_show_dictionary ? keyb_height [size] : keyb_top_height [size]);
	move_keyboard (current_profile.keyboard.x, current_profile.keyboard.y);

	/* Calculate the position of the keys */
	calculate_text_positions (&keyboard_lowercase);
	calculate_text_positions (&keyboard_uppercase);
	calculate_text_positions (&keyboard_numbers);
	calculate_text_positions (&keyboard_symbols_1);
	calculate_text_positions (&keyboard_symbols_2);
	calculate_text_positions (&keyboard_system);

	/* Of the dictionary buttons */
	x = x_dict_item [size];
	y = y_dict_item [size];
	j = 0;

	for (i = 0; i < DICT_ROWS * DICT_COLS; i++)
	{
		x_dict [i] = x + 10;
		y_dict [i] = y + (dict_height [size] / 2);

		x += dict_x_spacing [size];
		j++;

		if (j == DICT_COLS)
		{
			j = 0;
			x = x_dict_item [size];
			y += dict_y_spacing [size];
		}
	}

	/* And of the layout keys */
	x = keyb_width [size] + layout_bar_border [size];
	y = layout_bar_border [size];

	for (i = 0; i < N_KEYB_LAYOUTS + 1; i++)
	{
		x_layout_key [i] = x;
		y_layout_key [i] = y;

		y += layout_key_y_spacing [size];
	}

	gtk_widget_queue_draw (window);
}

	void
keyboard_handle_stopped (void)
{
	gint	n;

#ifdef __MACH__
	gint	x, y;
	gdk_window_get_pointer (window->window, &x, &y, NULL);
	pointer_at_position (x, y);
#endif	/* __MACH__ */

	/* Don't do anything until the pointer moves for the first time */
	if (first_time)
		return;

	/* If over the title bar, drag or drop */
	if (mouse_over_bar)
	{
		execute_action (dragging_kb ? DROP : DRAG);
		dragging_kb = !dragging_kb;

		if (dragging_kb)
			set_pointer (window, GDK_FLEUR);
		else
			gdk_window_set_cursor (window->window, NULL);

		gtk_widget_queue_draw (window);

		return;
	}

	/* If we are using switches, don't do anything */
	if (current_profile.config.use_switches)
		return;

	/* If we are over an empty dictionary button return */
	if (selected_col < KEYB_COLS && selected_row >= KEYB_ROWS)
	{
		n = (selected_row - KEYB_ROWS) * DICT_COLS + selected_col;

		if (!dictionary [n] && add_to_dict_entry != n)
			return;
	}

	if (selected_row < 0 || selected_col < 0)
		return;

	/* If the key is not active activate it. If it is, execute it. */
	if (!active)
	{
		if (selected_col < KEYB_COLS)
			active = TRUE;
		else
		{
			/* Except if it is a layout key, and the highlighted layout is
			 * already set... */
			if (selected_row != current_keyb_layout)
				active = TRUE;
		}
	}
	else
		execute_selected_key ();

	gtk_widget_queue_draw (window);
}

	void
keyboard_handle_switch (void)
{
	execute_selected_key ();
}

	static gboolean
window_expose_event_cb (GtkWidget*widget, GdkEventExpose *event, gpointer data)
{
	cairo_t*			cr;
	cairo_pattern_t*	pat;
	gint				i;
	gint				y;
	gint				selected_n;
	gchar				add_to_dict_text [WRITTEN_TEXT_BUFFER_SIZE + 2];

	cr = gdk_cairo_create (window->window);

	/* Layout bar background */
	pat = cairo_pattern_create_linear (keyb_width [size] - layout_bar_offset [size], 0,
			keyb_width [size] + layout_bar_width + layout_bar_offset [size], 0);

	cairo_pattern_add_color_stop_rgb (pat, 0.0, 0.6, 0.6, 0.6);
	cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.8, 0.8, 0.8);

	cairo_set_source (cr, pat);
	cairo_rectangle (cr, keyb_width [size] - layout_bar_offset [size], 0,
			layout_bar_width + layout_bar_offset [size], widget->allocation.height);
	cairo_fill (cr);

	cairo_pattern_destroy (pat);

	/* Layout bar border */
	cairo_set_line_width (cr, 1.0);
	cairo_set_source_rgb (cr, 0.6, 0.6, 0.6);

	cairo_rectangle (cr, keyb_width [size] - layout_bar_offset [size] + 0.5, 0.5,
			layout_bar_width + layout_bar_offset [size] - 1, widget->allocation.height - 1);
	cairo_stroke (cr);

	/* Dictionary background */
	cairo_set_source_rgb (cr, 0.91, 0.91, 0.91);
	cairo_rectangle (cr, 0, 0, keyb_width [size], keyb_height [size]);
	cairo_fill (cr);

	/* Dictionary border */
	cairo_set_source_rgb (cr, 0.6, 0.6, 0.6);
	cairo_set_line_width (cr, 1.0);

	cairo_rectangle (cr, 0.5, 0.5, keyb_width [size] - 1, keyb_height [size] - 1);
	cairo_stroke (cr);

	/* Keyboard background */
	cairo_set_source_rgb (cr, 0.875, 0.915, 0.96);
	cairo_rectangle (cr, 0, 0, keyb_width [size], keyb_top_height [size]);
	cairo_fill (cr);

	/* Keyboard border */
	cairo_set_source_rgb (cr, 0.6, 0.6, 0.6);
	cairo_set_line_width (cr, 1.0);

	cairo_rectangle (cr, 0.5, 0.5, keyb_width [size] - 1, keyb_top_height [size] - 1);
	cairo_stroke (cr);

	/* Title bar */
	edecor_draw_window_title_bar (cr, keyb_width [size], BAR_HEIGHT, mouse_over_bar);

	/* 
	 * Regular keys
	 */

	/* Find the selected key number */
	if (selected_col < KEYB_COLS)
	{
		if (selected_row < KEYB_ROWS - 1)
		{
			selected_n = selected_row * current_keyboard->columns +
				(selected_col - current_keyboard->offset_cols);
		}
		else if (selected_row >= KEYB_ROWS)
		{
			selected_n = (selected_row - KEYB_ROWS) * DICT_COLS +
				(selected_col - current_keyboard->offset_cols);
		}
		else
			selected_n = -1;
	}
	else
		selected_n = -1;

	for (i = 0; i < (KEYB_ROWS - 1) * current_keyboard->columns; i++)
	{
		/* Draw the button */
		draw_key (cr, current_keyboard->x_key [i], current_keyboard->y_key [i],	key_width [size],
				selected_row < KEYB_ROWS && selected_n == i);

		/* Write the text or draw the image */
		if (current_keyboard->key [i])
		{
			if (current_keyboard->key [i]->img)
			{
				cairo_set_source_surface (cr, current_keyboard->key [i]->img,
						current_keyboard->x_key [i], current_keyboard->y_key [i]);
				cairo_paint (cr);
			}
			else
			{
				cairo_set_source_rgb (cr, KEY_COLOR, KEY_COLOR, KEY_COLOR);
				edecor_draw_text_unformatted (cr, current_keyboard->x_text [i], current_keyboard->y_text [i], 
						current_keyboard->key [i]->text, 
						current_keyboard == &keyboard_system ? font_system [size] : font_key [size], TRUE);
			}
		}
	}

	/* Special keys */
	for (i = 0; i < N_SPECIAL_KEYS; i++)
	{
		if (!special_keys [i].visible)
			continue;

		draw_key (cr,
				x_keys [size] + (special_keys [i].col * keys_x_spacing [size]),
				y_keys [size] + (special_keys [i].row * keys_y_spacing [size]),
				key_width [size],
				selected_row == special_keys [i].row && selected_col == special_keys [i].col);

		cairo_set_source_surface (cr, special_keys [i].img,
				x_keys [size] + (special_keys [i].col * keys_x_spacing [size]),
				y_keys [size] + (special_keys [i].row * keys_y_spacing [size]));

		cairo_paint (cr);
	}

	/* Spacebar */
	draw_key (cr, x_keys [size] + (3 * keys_x_spacing [size]), y_keys [size] + (3 * keys_y_spacing [size]),
			3 * keys_x_spacing [size] + key_width [size],
			selected_row == 3 && selected_col > 2 && selected_col < 7);

	/* Dictionary */
	for (i = 0; i < DICT_ROWS * DICT_COLS; i++)
	{
		draw_dict_entry (cr, x_dict [i] - dict_x_text_offset [size], y_dict [i] - dict_y_text_offset [size], 
				selected_row >= KEYB_ROWS && selected_n == i);

		if (add_to_dict_entry == i)
		{
			strcpy (add_to_dict_text, "+ ");
			strcat (add_to_dict_text, written_text);

			cairo_set_source_rgb (cr, DICT_COLOR, DICT_COLOR, DICT_COLOR);
			edecor_draw_text_unformatted (cr, x_dict [i], y_dict [i], add_to_dict_text, font_dict [size], FALSE);
		}
		else if (dictionary [i] != NULL)
		{
			cairo_set_source_rgb (cr, DICT_COLOR, DICT_COLOR, DICT_COLOR);
			edecor_draw_text_unformatted (cr, x_dict [i], y_dict [i], dictionary [i], font_dict [size], FALSE);
		}
	}

	/* Layout keys */
	for (i = 0; i < N_KEYB_LAYOUTS + 1; i++)
	{
		draw_layout_key (cr, x_layout_key [i], y_layout_key [i], current_keyb_layout == i,
				selected_col == (KEYB_COLS + 1) && selected_row == i);

		if (i == N_KEYB_LAYOUTS)
		{
			cairo_set_source_surface (cr, images.os_layout,
					x_layout_key [i] + 2, y_layout_key [i] - 1);
			cairo_paint (cr);
		}
		else
		{
			cairo_set_source_rgb (cr, 0.3, 0.3, 0.3);
			edecor_draw_text_unformatted (cr, 
					x_layout_key [i] + (layout_key_width [size] >> 1) - 2,
					y_layout_key [i] + (layout_key_height [size] >> 1) - 1,
					keyb_layout_print_name [i], 
					current_keyb_layout == i ? font_layout_key_act [size] : font_layout_key [size], TRUE);
		}
	}

	/* System leds */
	draw_led (cr, "Ctrl", x_led_ctrl [size], y_leds [size], modifiers & MODIF_LEFT_CTRL);
	draw_led (cr, "Shift", x_led_shift [size], y_leds [size], modifiers & MODIF_LEFT_SHIFT);
	draw_led (cr, "Alt", x_led_alt [size], y_leds [size], modifiers & MODIF_LEFT_ALT);

	cairo_destroy (cr);

	return FALSE;
}

	static gboolean
window_motion_event_cb (GtkWidget *widget, GdkEventMotion *event, gpointer data)
{
	pointer_at_position ((gint) event->x, (gint) event->y);

	return FALSE;
}

	static gboolean
window_button_cb (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	if (event->button == 1 && event->type == GDK_BUTTON_PRESS &&
			event->y < BAR_HEIGHT) 
	{
		gtk_window_begin_move_drag (GTK_WINDOW (widget),
				event->button, event->x_root, event->y_root, event->time);
	}
}

	static gboolean
window_enter_event_cb (GtkWidget *widget, GdkEventCrossing *event, gpointer data)
{
	/* Mouse is over the keyboard */

	mouse_over_keyboard = TRUE;
	send_cfg_time ();

	gtk_widget_queue_draw (window);

	return FALSE;
}

	static gboolean
window_leave_event_cb (GtkWidget *widget, GdkEventCrossing *event, gpointer data)
{
	/* Mouse is away from keyboard */

	mouse_over_keyboard = FALSE;
	send_cfg_time ();

	mouse_over_bar = FALSE;

	gtk_widget_queue_draw (window);

	return FALSE;
}

	static gboolean
window_configure_event_cb (GtkWidget *widget, GdkEventConfigure *event, gpointer data)
{
	current_profile.keyboard.x = event->x;
	current_profile.keyboard.y = event->y;

	return FALSE;
}

	static void
pointer_at_position (gint x, gint y)
{
	gint 		movement_col, movement_row;
	gint 		left_col, right_col;
	gboolean	over_bar;
	
	over_bar = FALSE;

	first_time = FALSE;

	/* Where is the mouse? */

	if (x < keyb_width [size])
	{
		if (y < BAR_HEIGHT && x < keyb_width [size] - layout_bar_offset [size])
		{
			/* Over the bar */
			movement_row = -1;
			movement_col = -1;

			over_bar = TRUE;
		}
		else if (y > BAR_HEIGHT && y < y_dict_item [size])
		{
			/* Over the keys */
			if (x < x_keys [size])
				movement_col = 0;
			else
				movement_col = (x - x_keys [size]) / keys_x_spacing [size];

			if (y < y_keys [size])
				movement_row = 0;
			else
				movement_row = (y - y_keys [size]) / keys_y_spacing [size];

			if (movement_row > KEYB_ROWS - 1)
				movement_row = KEYB_ROWS - 1;

			if (movement_row == KEYB_ROWS - 1)
			{
				left_col = 0;
				right_col = KEYB_COLS - 1;
			}
			else
			{
				left_col = current_keyboard->offset_cols;
				right_col = current_keyboard->offset_cols + current_keyboard->columns - 1;
			}

			if (movement_col < left_col)
				movement_col = left_col;
			else if (movement_col > right_col)
				movement_col = right_col;
		}
		else if (y > BAR_HEIGHT)
		{
			/* Over the dictionary */
			if (x < x_dict_item [size])
				movement_col = 0;
			else
				movement_col = (x - x_dict_item [size]) / dict_x_spacing [size];

			movement_row = KEYB_ROWS + ((y - y_dict_item [size]) / dict_y_spacing [size]);

			if (movement_col > DICT_COLS - 1)
				movement_col = DICT_COLS - 1;
			if (movement_row > KEYB_ROWS + DICT_ROWS - 1)
				movement_row = KEYB_ROWS + DICT_ROWS - 1;
		}
	}
	else
	{
		/* Layout bar */
		movement_row = (y - layout_bar_border [size]) / layout_key_y_spacing [size];
		movement_col = KEYB_COLS + 1;
	}

	/* If over the title bar, configure a slower stop time */
	if (over_bar != mouse_over_bar)
	{
		mouse_over_bar = over_bar;

		if (mouse_over_bar)
			send_command (COMMAND_CONFIG, CFG_STOP_TIME, 150);
		else
			send_cfg_time ();

		gtk_widget_queue_draw (window);
	}

	/* The selected key is the one under the mouse */
	if (movement_col != selected_col || movement_row != selected_row)
	{
		selected_col = movement_col;
		selected_row = movement_row;

		active = FALSE;

		gtk_widget_queue_draw (window);
	}
}

	static void
move_keyboard (gint x, gint y)
{
	if (x == G_MAXINT)
		x = (screen_width - keyb_width [size]) >> 1;
	if (y == G_MAXINT)
		y = (screen_height - keyb_height [size] - 50);

	if (x < 0)
		x = 0;
	else if (x > screen_width - keyb_width [size])
		x = screen_width - keyb_width [size];

	if (y < 0)
		y = 0;
	else if (y > screen_height - keyb_height [size])
		y = screen_height - keyb_height [size];

	gtk_window_move (GTK_WINDOW (window), x, y);
}

	static void
type_key (Key* key)
{
	/* If a dead key is to be typed before, type it */
	if (key->dead_key)
		type_key (key->dead_key);

	/* Send the key code and modifiers */
	send_command (COMMAND_KEY, key->code, key->modifiers | modifiers);
}

	static void
draw_key (cairo_t* cr, gfloat x, gfloat y, gfloat width, gboolean selected)
{
	cairo_pattern_t*	pat;

	/* Background pattern */
	pat = cairo_pattern_create_linear (x, y, x, y + key_height [size]);

	if (selected && mouse_over_keyboard)
	{
		if (active)
		{
			cairo_pattern_add_color_stop_rgb (pat, 0.0, 0.75, 0.75, 0.75);
			cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.9, 0.9, 0.9);
		}
		else
		{
#ifndef __MACH__
			cairo_pattern_add_color_stop_rgb (pat, 0.0, 0.99, 0.99, 0.99);
			cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.91, 0.91, 0.91);
#else
			cairo_pattern_add_color_stop_rgb (pat, 0.0, 0.95, 0.95, 0.95);
			cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.87, 0.87, 0.87);
#endif	/* __MACH__ */
		}
	}
	else
	{
		cairo_pattern_add_color_stop_rgb (pat, 0.0, 0.95, 0.95, 0.95);
		cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.87, 0.87, 0.87);
	}

	edecor_draw_round_rect (cr, x, y, width, key_height [size], radius_key_border [size]);

	cairo_set_source (cr, pat);
	cairo_fill (cr);

	cairo_pattern_destroy (pat);

	/* Border */
	cairo_set_line_width (cr, 0.5);
	cairo_set_source_rgb (cr, 0.6, 0.6, 0.6);

	edecor_draw_round_rect (cr, x, y, width, key_height [size], radius_key_border [size]);

	/* Highlighted border */
	if (selected && mouse_over_keyboard)
	{
#ifndef __MACH__
		cairo_set_line_width (cr, 1.0);

		if (active)
			cairo_set_source_rgb (cr, 0.914, 0.192, 0.192);
		else
			cairo_set_source_rgb (cr, 0.18, 0.3, 0.55);

		edecor_draw_round_rect (cr, x + 1, y + 1, width - 2, key_height [size] - 2, radius_key_border [size] - 1);
#else
		if (active)
		{
			cairo_set_source_rgb (cr, 0.914, 0.192, 0.192);
			edecor_draw_round_rect (cr, x + 1, y + 1, width - 2, key_height [size] - 2, radius_key_border [size] - 1);
		}
#endif	/* __MACH__ */
	}

	cairo_stroke (cr);
}

	static void
draw_dict_entry (cairo_t* cr, gfloat x, gfloat y, gboolean selected)
{
	cairo_pattern_t*	pat;

	/* Background pattern */
	pat = cairo_pattern_create_linear (x, y, x, y + dict_height [size]);

	if (selected && mouse_over_keyboard)
	{
		if (active)
		{
			cairo_pattern_add_color_stop_rgb (pat, 0.0, 0.75, 0.75, 0.75);
			cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.9, 0.9, 0.9);
		}
		else
		{
			cairo_pattern_add_color_stop_rgb (pat, 0.0, 0.99, 0.99, 0.99);
			cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.91, 0.91, 0.91);
		}
	}
	else
	{
		cairo_pattern_add_color_stop_rgb (pat, 0.0, 0.95, 0.95, 0.95);
		cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.87, 0.87, 0.87);
	}

	edecor_draw_round_rect (cr, x, y, dict_width [size] - 4, dict_height [size] - 3, radius_dict_border [size]);

	cairo_set_source (cr, pat);
	cairo_fill (cr);

	cairo_pattern_destroy (pat);

	/* Border */
	cairo_set_line_width (cr, 0.5);
	cairo_set_source_rgb (cr, 0.6, 0.6, 0.6);

	edecor_draw_round_rect (cr, x, y, dict_width [size] - 4, dict_height [size] - 3, radius_dict_border [size]);

	/* Highlighted border */
	if (selected && mouse_over_keyboard)
	{
		cairo_set_line_width (cr, 1.0);

		if (active)
			cairo_set_source_rgb (cr, 0.914, 0.192, 0.192);
		else
			cairo_set_source_rgb (cr, 0.18, 0.3, 0.55);

		edecor_draw_round_rect (cr, x + 1, y + 1, dict_width [size] - 6, dict_height [size] - 5,
				radius_dict_border [size] - 1);
	}

	cairo_stroke (cr);
}

	static void
draw_layout_key (cairo_t* cr, gfloat x, gfloat y, gboolean current, gboolean selected)
{
	cairo_pattern_t*	pat;

	/* Background pattern */
	pat = cairo_pattern_create_linear (x, y, x, y + dict_height [size]);

	if (current)
	{
		cairo_pattern_add_color_stop_rgb (pat, 0.0, 0.65, 0.65, 0.65);
		cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.8, 0.8, 0.8);
	}
	else if (selected && mouse_over_keyboard)
	{
		if (active)
		{
			cairo_pattern_add_color_stop_rgb (pat, 0.0, 0.65, 0.65, 0.65);
			cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.8, 0.8, 0.8);
		}
		else
		{
			cairo_pattern_add_color_stop_rgb (pat, 0.0, 0.99, 0.99, 0.99);
			cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.91, 0.91, 0.91);
		}
	}
	else
	{
		cairo_pattern_add_color_stop_rgb (pat, 0.0, 0.95, 0.95, 0.95);
		cairo_pattern_add_color_stop_rgb (pat, 1.0, 0.87, 0.87, 0.87);
	}

	edecor_draw_round_rect (cr, x, y, layout_key_width [size], layout_key_height [size],
			radius_layout_bar_border [size]);

	cairo_set_source (cr, pat);
	cairo_fill (cr);

	cairo_pattern_destroy (pat);

	/* Border */
	cairo_set_line_width (cr, 1.0);
	cairo_set_source_rgb (cr, 0.6, 0.6, 0.6);

	edecor_draw_round_rect (cr, x + 0.5, y + 0.5,
			layout_key_width [size] - 1, layout_key_height [size] - 1, radius_layout_bar_border [size]);

	/* Highlighted border */
	if (selected && mouse_over_keyboard)
	{
		cairo_set_line_width (cr, 1.0);

		if (active)
			cairo_set_source_rgb (cr, 0.914, 0.192, 0.192);
		else
			cairo_set_source_rgb (cr, 0.18, 0.3, 0.55);

		edecor_draw_round_rect (cr, x + 1.5, y + 1.5,
				layout_key_width [size] - 3, layout_key_height [size] - 3,
				radius_layout_bar_border [size] - 1.5);
	}

	cairo_stroke (cr);
}

	static void
draw_led (cairo_t* cr, gchar* text, gfloat x, gfloat y, gboolean on)
{
	cairo_set_source_rgb (cr, 0.6, 0.6, 0.6);
	edecor_draw_text_unformatted (cr, x + radius_led [size] + 2, y, text, font_led [size], FALSE);

	/* Background */
	if (on)
		cairo_set_source_rgb (cr, 0.192, 0.678, 0.914);
	else
		cairo_set_source_rgb (cr, 0.7, 0.7, 0.7);

	cairo_arc (cr, x, y, radius_led [size], 0, 2.0 * M_PI);
	cairo_fill (cr);

	/* Border */
	cairo_set_line_width (cr, 0.5);
	cairo_set_source_rgb (cr, 0.6, 0.6, 0.6);

	cairo_arc (cr, x, y, radius_led [size], 0, 2.0 * M_PI);
	cairo_stroke (cr);
}

	static void
load_keyboards (void)
{
	gchar	keyb_dir [200];
	gchar	path [300];
	gchar*	data_dir;

	data_dir = get_data_file ("/keyboards/");
	sprintf (keyb_dir, "%s%s", data_dir, keyb_layout_name [current_keyb_layout]);
	g_free (data_dir);

	loaded_keyb_layout = current_keyb_layout;

	/* Load the layout */
	sprintf (path, "%s/layout", keyb_dir);
	load_layout (&keyboard_layout, path);

	/* And the keyboards */
	sprintf (path, "%s/lowercase", keyb_dir);
	load_keyboard (&keyboard_lowercase, keyboard_layout, path);
	sprintf (path, "%s/uppercase", keyb_dir);
	load_keyboard (&keyboard_uppercase, keyboard_layout, path);
	sprintf (path, "%s/numbers", keyb_dir);
	load_keyboard (&keyboard_numbers,	keyboard_layout, path);
	sprintf (path, "%s/symbols1", keyb_dir);
	load_keyboard (&keyboard_symbols_1, keyboard_layout, path);
	sprintf (path, "%s/symbols2", keyb_dir);
	load_keyboard (&keyboard_symbols_2,	keyboard_layout, path);
	sprintf (path, "%s/system", keyb_dir);
	load_keyboard (&keyboard_system,	keyboard_layout, path);
}

	static gint
load_layout (GSequence** layout, gchar* file)
{
	FILE*			f;
	gchar			line [100];
	gchar*			token;
	guint			code;
	gchar*			id;
	gchar*			text;
	gchar*			modifiers;
	gchar*			dead_key_id;
	Key*			key;
	Key*			dead_key;
	GSequenceIter*	iter;

	/* Create the sequence */
	*layout = g_sequence_new (free_key);

	/* Open the file */
	f = (FILE*) g_fopen (file, "r");
	if (!f)
		return -1;

	/* Read the keys */
	while (!feof (f))
	{
		/* Read a line */
		if (fgets (line, sizeof (line), f) == NULL)
			continue;

		/* Read the code */
		token = strtok (line, " \t");
		sscanf (token, "%x", &code);

		/* Read the keys */
		while (token)
		{
			/* Get the id */
			id = NULL;

			if (*token == '[')
			{
				token++;
				id = token;

				token++;
				while (*token && *token != ']')
					token++;

				*token = 0;
				token++;
			}

			if (id != NULL && *id != 0)
			{
				/* Allocate a key */
				key = g_malloc (sizeof (Key));

				/* Get the text, modifiers and dead keys */
				text = NULL;
				modifiers = NULL;
				dead_key_id = NULL;

				while (*token)
				{
					/* Text */
					if (*token == '"')
					{
						token++;
						text = token;

						while (*token && *token != '"')
							token++;

						*token = 0;
						token++;
					}
					/* Modifiers */
					else if (*token == '(')
					{
						token++;
						modifiers = token;

						while (*token && *token != ')')
							token++;

						*token = 0;
						token++;
					}
					/* Dead keys */
					else if (*token == '{')
					{
						token++;
						dead_key_id = token;

						while (*token && *token != '}')
							token++;

						*token = 0;
						token++;
					}
					else
						token++;
				}

				key->id = g_strdup (id);
				key->text = (text ? g_strdup (text) : g_strdup (id));
				key->code = (guchar) code;

				key->modifiers = 0;
				if (modifiers != NULL)
				{
					while (*modifiers)
					{
						switch (*modifiers)
						{
							case 'S':
								key->modifiers |= MODIF_LEFT_SHIFT;	break;
							case 'G':
								key->modifiers |= MODIF_RIGHT_ALT;	break;
							case 'C':
								key->modifiers |= MODIF_LEFT_CTRL;	break;
						}

						modifiers++;
					}
				}

				key->dead_key_id = g_strdup (dead_key_id);

				/* Set the image */
				key->img = NULL;
				switch (key->code)
				{
					case KEY_ARROW_UP:
						key->img = images.up;
						break;
					case KEY_ARROW_DOWN:
						key->img = images.down;
						break;
					case KEY_ARROW_LEFT:
						key->img = images.left;
						break;
					case KEY_ARROW_RIGHT:
						key->img = images.right;
						break;
					case KEY_APLICATION:
						key->img = images.options;
						break;
					case KEY_ESCAPE:
						if (key->modifiers == MODIF_LEFT_CTRL)
							key->img = images.operating_system;
						break;
				}

				g_sequence_append (*layout, key);
			}

			token = strtok (NULL, " \t");
		}
	}

	/* Find the codes for the dead keys */
	iter = g_sequence_get_begin_iter (*layout);
	while (iter != g_sequence_get_end_iter (*layout))
	{
		key = g_sequence_get (iter);

		if (key->dead_key_id != NULL)
		{
			dead_key = find_key (key->dead_key_id, *layout);

			if (dead_key != NULL)
				key->dead_key = dead_key;
		}
		else
			key->dead_key = NULL;

		iter = g_sequence_iter_next (iter);
	}

	fclose (f);

	return 0;
}

	static gint
load_keyboard (Keyboard* keyboard, GSequence* layout, gchar* file)
{
	cairo_surface_t* key_img;
	FILE*			f;
	gchar			line [50];
	gchar*			id;
	gint			n_keys, i;

	/* Open the file */
	f = (FILE*) g_fopen (file, "r");
	if (!f)
		return -1;

	/* Read the columns count and the offset */
	if (fscanf (f, "%d", &keyboard->columns) <= 0)
		return -1;
	if (fscanf (f, "%d", &keyboard->offset_cols) <= 0)
		return -1;

	/* Read the keyboard */
	n_keys = (KEYB_ROWS - 1) * keyboard->columns;

	i = 0;
	while (!feof (f))
	{
		keyboard->key [i] = NULL;

		/* Read a line */
		if (fgets (line, sizeof (line), f) == NULL)
			continue;

		/* Get the id */
		id = strtok (line, "\n\r ");
		if (id == NULL)
			continue;

		/* Find the id in the layout */
		keyboard->key [i] = find_key (id, layout);

		if (keyboard->key [i] == NULL)
			g_warning ("Key id not found in layout: '%s'", id);

		i++;
		if (i == n_keys)
			break;
	}

	fclose (f);

	return 0;
}

	static void
calculate_text_positions (Keyboard* keyboard)
{
	gfloat	x0, x, y;
	gint	n_keys, i, j;

	/* The position of the texts depend on the number of columns, 
	 * the offset, the keyboard size, etc. */

	x0 = x_keys [size] + (keyboard->offset_cols * keys_x_spacing [size]);
	x = x0;
	y = y_keys [size];
	j = 0;

	n_keys = (KEYB_ROWS - 1) * keyboard->columns;

	for (i = 0; i < n_keys; i++)
	{
		keyboard->x_text [i] = x + (key_width [size] / 2);
		keyboard->y_text [i] = y + (key_height [size] / 2);

		keyboard->x_key [i] = x;
		keyboard->y_key [i] = y;

		x += keys_x_spacing [size];
		j++;

		if (j == keyboard->columns)
		{
			j = 0;
			y += keys_y_spacing [size];
			x = x0;
		}
	}
}

	static void
free_key (gpointer data)
{
	Key*	key;

	key = (Key*) data;

	g_free (key->id);
	g_free (key->text);
	g_free (key->dead_key_id);

	g_free (key);
}

	static Key*
find_key (gchar* id, GSequence* layout)
{
	Key*			key;
	GSequenceIter*	iter;

	iter = g_sequence_get_begin_iter (layout);
	while (iter != g_sequence_get_end_iter (layout))
	{
		key = g_sequence_get (iter);

		if (!strcmp (key->id, id))
			return key;

		iter = g_sequence_iter_next (iter);
	}

	return NULL;
}

	static void
init_dictionary_entries (void)
{
	gint i;

	strcpy (written_text, "");
	written_text_n_chars = 0;

	words = NULL;

	for (i = 0; i < DICT_ROWS * DICT_COLS; i++)
		dictionary [i] = NULL;
}

	static void
load_dictionaries (void)
{
	gchar*	dicts_dir;
	gchar*	lang_dict_path;
	gchar*	user_dict_path;

	dicts_dir = get_data_file ("/dictionaries/");

	lang_dict_path = g_build_filename (dicts_dir,
			keyb_dictionary_name [current_keyb_layout], NULL);
	user_dict_path = g_build_filename (dicts_dir, "user", NULL);

	g_free (dicts_dir);

	init_dictionary ();
	load_dictionary (user_dict_path);
	load_dictionary (lang_dict_path);

	g_free (lang_dict_path);
	g_free (user_dict_path);
}

	static void
init_special_keys (void)
{
	special_keys [sk_numbers].visible = TRUE;
	special_keys [sk_numbers].row = 3;
	special_keys [sk_numbers].img = images.numbers;

	special_keys [sk_letters].visible = TRUE;
	special_keys [sk_letters].row = 3;
	special_keys [sk_letters].img = images.letters;

	special_keys [sk_symbols].visible = TRUE;
	special_keys [sk_symbols].row = 3;
	special_keys [sk_symbols].img = images.symbols;

	special_keys [sk_system].visible = TRUE;
	special_keys [sk_system].row = 3;
	special_keys [sk_system].img = images.system;

	special_keys [sk_backspace].visible = TRUE;
	special_keys [sk_backspace].row = 3;
	special_keys [sk_backspace].col = 7;
	special_keys [sk_backspace].img = images.backspace;

	special_keys [sk_return].visible = TRUE;
	special_keys [sk_return].row = 3;
	special_keys [sk_return].col = 8;
	special_keys [sk_return].img = images.return_key;

	special_keys [sk_close].visible = TRUE;
	special_keys [sk_close].row = 3;
	special_keys [sk_close].col = 9;
	special_keys [sk_close].img = images.close;

	special_keys [sk_uppercase].visible = TRUE;
	special_keys [sk_uppercase].row = 2;
	special_keys [sk_uppercase].col = 9;
	special_keys [sk_uppercase].img = images.uppercase;

	special_keys [sk_more_symbols].visible = TRUE;
	special_keys [sk_more_symbols].row = 2;
	special_keys [sk_more_symbols].col = 9;
	special_keys [sk_more_symbols].img = images.more_symbols;
}

	static void
change_current_keyboard (Keyboard* keyboard)
{
	gint i, col;

	current_keyboard = keyboard;

	/* The keyboard selection keys appear or disappear depending on
	 * the current keyboard */
	special_keys [sk_letters].visible = 
		keyboard != &keyboard_lowercase && keyboard != &keyboard_uppercase;
	special_keys [sk_numbers].visible = 
		keyboard != &keyboard_numbers;
	special_keys [sk_symbols].visible = 
		keyboard != &keyboard_symbols_1 && keyboard != &keyboard_symbols_2;
	special_keys [sk_system].visible = 
		keyboard != &keyboard_system;

	col = 0;
	for (i = 0; i < 4; i++)
	{
		if (special_keys [i].visible)
		{
			special_keys [i].col = col;
			col++;
		}
	}

	/* Uppercase/lowercase key */
	special_keys [sk_uppercase].visible =
		layout_has_uppercase [loaded_keyb_layout] &&
		(keyboard == &keyboard_lowercase || keyboard == &keyboard_uppercase);

	special_keys [sk_uppercase].img = 
		keyboard == &keyboard_uppercase ? images.lowercase : images.uppercase;

	/* More symbols key */
	special_keys [sk_more_symbols].visible =
		(keyboard == &keyboard_symbols_1 || keyboard == &keyboard_symbols_2) &&
		layout_has_symbols_2 [loaded_keyb_layout];
}

	static void
execute_selected_key (void)
{
	gint		n, i;
	gchar*		p;
	gchar		id [5];
	guchar		letter;
	gsize		s, t;
	Key*		key;
	gboolean	found;

	/* Check whether it is a special key */
	for (i = 0; i < N_SPECIAL_KEYS; i++)
	{
		if (special_keys [i].visible &&
				selected_row == special_keys [i].row &&
				selected_col == special_keys [i].col)
		{
			break;
		}
	}

	if (i < N_SPECIAL_KEYS)
	{
		/* A special key */
		switch (i)
		{
			case sk_numbers:
				/* Keyboard change */
				change_current_keyboard (&keyboard_numbers);
				break;
			case sk_letters:
				change_current_keyboard (&keyboard_lowercase);
				break;
			case sk_symbols:
				change_current_keyboard (&keyboard_symbols_1);
				break;
			case sk_system:
				change_current_keyboard (&keyboard_system);
				break;
			case sk_close:
				/* Close */
				hide_keyboard ();
				break;
			case sk_backspace:
				/* Backspace */
				if (written_text_n_chars > 0)
				{
					written_text_n_chars--;
					*(g_utf8_offset_to_pointer (written_text, written_text_n_chars)) = 0;
				}
				send_command (COMMAND_KEY, KEY_BACKSPACE, 0);
				break;
			case sk_return:
				/* Return */
				written_text [0] = 0;
				written_text_n_chars = 0;
				send_command (COMMAND_KEY, KEY_ENTER, 0);
				break;
			case sk_uppercase:
				/* Uppercase/lowercase */
				change_current_keyboard (
						current_keyboard == &keyboard_lowercase ?
						&keyboard_uppercase : &keyboard_lowercase);
				break;
			case sk_more_symbols:
				/* More symbols */
				change_current_keyboard (
						current_keyboard == &keyboard_symbols_1 ?
						&keyboard_symbols_2 : &keyboard_symbols_1);
				break;
			default: 
				break;
		}
	}
	else if (selected_col == KEYB_COLS + 1)
	{
		/* A layout key */
		if (selected_row == N_KEYB_LAYOUTS)
		{
#ifdef G_OS_WIN32
			KEYBDINPUT	kb	= { 0 };
			INPUT		input;

			input.type = INPUT_KEYBOARD;

			/* In Windows, Alt + Left shift switches the input device layout */

			/* Left shift down */
			kb.wVk = VK_LSHIFT;
			input.ki = kb;
			SendInput (1, &input, sizeof (input));

			/* Alt down */
			kb.wVk = VK_MENU;
			input.ki = kb;
			SendInput (1, &input, sizeof (input));

			/* Left shift up */
			kb.wVk = VK_LSHIFT;
			kb.dwFlags |= KEYEVENTF_KEYUP;
			input.ki = kb;
			SendInput (1, &input, sizeof (input));

			/* Alt up */
			kb.wVk = VK_MENU;
			kb.dwFlags |= KEYEVENTF_KEYUP;
			input.ki = kb;
			SendInput (1, &input, sizeof (input));
#endif
		}
		else
		{
			current_keyb_layout = selected_row;

			update_config_file ();
			update_main_window ();

			reset_keyboard ();
		}
	}
	else
	{
		if (selected_row == KEYB_ROWS - 1)
		{
			/* Spacebar */
			strcpy (written_text, "");
			send_command (COMMAND_KEY, KEY_SPACE, 0);
		}
		else if (selected_row < KEYB_ROWS - 1)
		{
			/* Some other key */
			i = selected_row * current_keyboard->columns + 
				(selected_col - current_keyboard->offset_cols);

			switch (current_keyboard->key [i]->code)
			{
				/* If it is a modifier, (de)activate it */
				case KEY_CTRL_LEFT:
					modifiers ^= MODIF_LEFT_CTRL;
					break;
				case KEY_SHIFT_LEFT:
					modifiers ^= MODIF_LEFT_SHIFT;
					break;
				case KEY_ALT_LEFT:
					modifiers ^= MODIF_LEFT_ALT;
					break;

					/* A regular letter */
				default:

					type_key (current_keyboard->key [i]);

					/* If in uppercase, return to lowercase */
					if (current_keyboard == &keyboard_uppercase)
						change_current_keyboard (&keyboard_lowercase);

					/* Look up in the dictionary the text written so far */
					if (current_keyboard->key [i]->code == KEY_ENTER &&
							current_keyboard->key [i]->modifiers == 0) strcpy (written_text, "");
					else
					{
						if (strlen (written_text) + strlen (current_keyboard->key [i]->text) <
								sizeof (written_text))
						{
							strcat (written_text, current_keyboard->key [i]->text);
						}
					}

					break;
			}
		}
		else
		{
			/* A dictionary word */
			n = (selected_row - KEYB_ROWS) * DICT_COLS + selected_col;

			if (n == add_to_dict_entry)
			{
				gchar	text [WRITTEN_TEXT_BUFFER_SIZE + 20];
				gchar*	dicts_dir;
				gchar*	user_dict_path;
				FILE*	f;

				/* Get the dictionary path */
				dicts_dir = get_data_file ("/dictionaries/");

				user_dict_path = g_build_filename (dicts_dir, "user", NULL);

				g_free (dicts_dir);

				/* Append the new word */
				f = (FILE*) g_fopen (user_dict_path, "a");

				if (f)
				{
					fwrite (written_text, 1, strlen (written_text), f);
					fwrite ("\n", 1, strlen ("\n"), f);

					fclose (f);
				}

				g_free (user_dict_path);

				/* Show an info dialog */
				sprintf (text, "%s: '%s'", _("Added"), written_text);
				dialog_transient (text, "images/enabled.png");

				/* Reload the dictionaries */
				load_dictionaries ();
			}
			else
			{
				p = dictionary [n];

				if (p)
				{
					strcpy (written_text, p);
					for (i = 0; i < written_text_n_chars; i++)
						p = g_utf8_next_char (p);

					while (*p)
					{
						/* Find the letter in the layout and type it */
						g_utf8_strncpy (id, p, 1);

						key = find_key (id, keyboard_layout);
						if (key != NULL)
							type_key (key);

						g_usleep (100000);

						p = g_utf8_next_char (p);
					}
				}
			}
		}	

		written_text_n_chars = g_utf8_strlen (written_text, -1);
	}

	/* Find words that start with the written text */
	if (words)
	{
		g_string_free (words, TRUE);
		words = NULL;
	}

	/* Clear the dictionary entries */
	for (i = 0; i < DICT_ROWS * DICT_COLS; i++)
		dictionary [i] = NULL;

	look_up_dictionary (written_text, &words);

	n = DICT_ROWS * DICT_COLS;
	found = FALSE;

	if (words != NULL)
	{
		p = words->str;
		s = 0;
		i = 0;

		/* Fill in the entries */
		while (s < words->len && n > 0)
		{
			if (strcmp (written_text, p))
			{
				dictionary [i] = p;
				n--;
				i++;
			}
			else
				found = TRUE;

			t = strlen (p) + 1;

			p += t;
			s += t;
		}

		if (!found && n == 0)
			n++;
	}

	/* If the written text is not in the dictionary, add an
	 * option to add it */
	if (found || strlen (written_text) < 3)
		add_to_dict_entry = -1;
	else
		add_to_dict_entry = (DICT_ROWS * DICT_COLS) - n;

	active = FALSE;
}
