/* 
 * enPathia
 * Eneso Tecnología de Adaptación, S.L. 
 */

#include <cairo.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <stdlib.h>
#include <string.h>
#include "edecor.h"
#include "action.h"
#include "calibration.h"
#include "enpathia.h"
#include "keyboard.h"
#include "misc.h"
#include "profile.h"
#include "protocol.h"
#include "unlock.h"
#include "menu.h"

#define NONE	-1
#define UP		0
#define RIGHT	1
#define LEFT	2
#define DOWN	3

#define MORE	DOWN

gboolean showing_menu	= FALSE;
gboolean menu_disabled	= FALSE;

/* Menus */
Menu menu_moving;
Menu menu_dragging;

static Menu* current_menu;

static gint selected_button;
static gint offset; 

static GtkWidget* window;

static gint x_button [4];
static gint y_button [4];

static const guint	width [N_MENU_SIZES]			= { 155, 310, 620 };
static const guint	height [N_MENU_SIZES]			= { 103, 206, 412 };
static const guint	bar_width [N_MENU_SIZES]		= { 50, 100, 200 };
static const gchar*	font [N_MENU_SIZES]				= { "Sans 11px", "Sans 23px", "Sans 45px" };

static gint size = 0;

static gchar button_more_text [20];

static gboolean
window_expose_cb	(GtkWidget *widget, GdkEventExpose *event, gpointer data);

static void
init_menus	();


	void
create_menu (void)
{
	/* Create the window */
	window = gtk_window_new (GTK_WINDOW_POPUP);

	g_signal_connect (window,
			"expose-event", G_CALLBACK (window_expose_cb), NULL);

	gtk_window_set_title (GTK_WINDOW (window), "enPathia");
	gtk_window_set_decorated (GTK_WINDOW (window), FALSE);
	gtk_widget_set_app_paintable (window, TRUE);
	gtk_window_set_keep_above (GTK_WINDOW (window), TRUE);		

	strcpy (button_more_text, _("More"));

#ifndef G_OS_WIN32
	GdkScreen*		screen;
	GdkColormap*	colormap;

	/* Set RGBA colormap */ 
	screen = gtk_widget_get_screen (window);
	colormap = gdk_screen_get_rgba_colormap (screen);

	gtk_widget_set_colormap (window, colormap);
#endif

	/* Initialize the menus */
	init_menus ();
}

	void
show_menu (gint x, gint y)
{
	if (showing_menu || calibrating || unlocking)
		return;

	showing_menu = TRUE;

	/* Change to notification mode */
	change_mode (MODE_NOTIFY);

	offset = 0;
	selected_button = NONE;

	/* Center the window where we are asked to and show it */
	x -= (width [size] >> 1);
	if (x < 0)
		x = 0;
	else if (x + width [size] > screen_width)
		x = screen_width - width [size];

	y -= (height [size] >> 1);
	if (y < 0)
		y = 0;
	else if (y + height [size] > screen_height)
		y = screen_height - height [size];

	gtk_window_move (GTK_WINDOW (window), x, y);
	gtk_window_present (GTK_WINDOW (window));

#ifdef __MACH__
	gtk_window_set_keep_above (GTK_WINDOW (window), TRUE);		
#endif
}

	void
hide_menu (void)
{
	showing_menu = FALSE;

	gtk_widget_hide (window);
}

	void
set_menu_size (gint menu_size)
{
	gint	i;

	if (menu_size < 0)
		menu_size = 0;
	else if (menu_size > N_MENU_SIZES - 1)
		menu_size = N_MENU_SIZES - 1;

	size = menu_size;

	gtk_window_resize (GTK_WINDOW (window), width [size], height [size]);

	/* Calculate the position of the buttons */
	x_button [UP] = x_button [DOWN] = (width [size] >> 1);
	x_button [LEFT] = (width [size] * 0.2);
	x_button [RIGHT] = (width [size] * 0.8);

	y_button [LEFT] = y_button [RIGHT] = (height [size] >> 1);
	y_button [UP] = (height [size] * 0.2);
	y_button [DOWN] = (height [size] * 0.8);
}

	void
change_menu (Menu* menu)
{
	current_menu = menu;

	gtk_widget_queue_draw (window);
}

	void
menu_handle_movement (gchar movement)
{
	/* Select the button at the given position */
	switch (movement)
	{
		case MOV_UP:
			selected_button = UP;
			break;

		case MOV_DOWN:
			selected_button = DOWN;
			break;

		case MOV_LEFT:
			selected_button = LEFT;
			break;

		case MOV_RIGHT:
			selected_button = RIGHT;
			break;
	}

	/* Empty buttons cannot be selected */
	if (selected_button == MORE)
	{
		if (current_menu->n_actions < 4)
			selected_button = NONE;
	}
	else
	{
		if (offset + selected_button >= current_menu->n_actions)
			selected_button = NONE;
	}

	gtk_widget_queue_draw (window);
}

	void
menu_handle_stopped (void)
{
	Action action;

	/* Execute the selected button. If it is 'More' increase the menu offset.
	 * If it is an action launch it */
	if (current_menu->n_actions > 4 && selected_button == MORE)
	{
		offset += 3;
		if (offset >= current_menu->n_actions)
			offset = 0;
	}
	else if (selected_button != NONE)
	{
		action = current_menu->actions [selected_button + offset];

		hide_menu ();

#ifdef G_OS_UNIX
#ifndef __MACH__
		/* Wait for the window under the pointer to grab the focus */
		wait_ms (100);
#endif
#endif

		execute_action (action);
	}

	selected_button = NONE;

	gtk_widget_queue_draw (window);
}

	void
set_menu_disabled (gboolean disabled)
{
	menu_disabled = disabled;

	dialog_transient (menu_disabled ? _("Menu disabled") : _("Menu enabled"),
			menu_disabled ? "images/disabled.png" : "images/enabled.png");
}

	static gboolean
window_expose_cb (GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
	cairo_t*			cr;
	cairo_surface_t*	image;
	gchar* 				text;
	Action 				action;
	gboolean			disabled;
	gint 				i;

	cr = gdk_cairo_create (widget->window);

	/* Buttons */
	cairo_set_line_width (cr, 1);

	for (i = 0; i < 4; i++)
	{
		text = NULL;
		image = NULL;
		disabled = FALSE;

		/* If it is an action get the associated text */
		if (i == DOWN && current_menu->n_actions >= 3)
		{
			text = button_more_text;
			image = action_more_image [size];
		}
		else if (i != DOWN && offset + i < current_menu->n_actions)
		{
			action = current_menu->actions [offset + i];
			text = action_text [action];
			image = action_image [action][size];
		}
		/* Otherwise, disable the button */
		else
			disabled = TRUE;

		/* Background */
		if (selected_button == i || disabled)
		{
			switch (i)
			{
				case UP:
					cairo_move_to (cr, 0, 0);
					cairo_line_to (cr, width [size], 0);
					cairo_line_to (cr, width [size] >> 1, height [size] >> 1);
					break;
				case RIGHT:
					cairo_move_to (cr, width [size], 0);
					cairo_line_to (cr, width [size], height [size]);
					cairo_line_to (cr, width [size] >> 1, height [size] >> 1);
					break;
				case DOWN:
					cairo_move_to (cr, 0, height [size]);
					cairo_line_to (cr, width [size], height [size]);
					cairo_line_to (cr, width [size] >> 1, height [size] >> 1);
					break;
				case LEFT:
					cairo_move_to (cr, 0, 0);
					cairo_line_to (cr, 0, height [size]);
					cairo_line_to (cr, width [size] >> 1, height [size] >> 1);
					break;
			}

			if (disabled)
				cairo_set_source_rgb (cr, 0.6, 0.6, 0.6);
			else if (selected_button == i)
				cairo_set_source_rgb (cr, 0.33, 0.51, 0.84);
			else
				cairo_set_source_rgb (cr, 0.95, 0.95, 0.95);

			cairo_fill (cr);
		}

		/* Write the text or draw the image */
		if (current_profile.config.menu_use_images)
		{
			if (image)
			{
				cairo_set_source_surface (cr, image,
						x_button [i] - (cairo_image_surface_get_width (image) >> 1),
						y_button [i] - (cairo_image_surface_get_height (image) >> 1));
				cairo_paint (cr);
			}
		}
		else
		{
			if (text)
			{
				if (selected_button == i)
					cairo_set_source_rgb (cr, 0.9, 0.9, 0.9);
				else
					cairo_set_source_rgb (cr, 0, 0, 0);

				edecor_draw_text (cr, x_button [i], y_button [i],
						text, font [size], TRUE);
			}
		}
	}

	/* Lines */
	cairo_set_source_rgb (cr, 0.8, 0.8, 0.8);
	cairo_set_line_width (cr, 1.0);

	cairo_move_to (cr, 0, 0);
	cairo_line_to (cr, width [size], height [size]);
	cairo_stroke (cr);

	cairo_move_to (cr, width [size], 0);
	cairo_line_to (cr, 0, height [size]);
	cairo_stroke (cr);

	/* Border */
	cairo_rectangle (cr, 0.5, 0.5, width [size] - 1, height [size] - 1);
	cairo_stroke (cr);

	cairo_destroy (cr);

	return FALSE;
}

	static void
init_menus ()
{
	/* Movement menu */
	menu_moving.n_actions = 6;
	menu_moving.actions = 
		(Action*) g_malloc (sizeof (Action) * menu_moving.n_actions);

	menu_moving.actions [0] = CANCEL;
	menu_moving.actions [1] = OPTIONS;
	menu_moving.actions [2] = CLICK;
	menu_moving.actions [3] = KEYBOARD;
	menu_moving.actions [4] = DRAG;
	menu_moving.actions [5] = EXECUTE;

	current_menu = &menu_moving;
	offset = 0;

	/* Dragging menu */
	menu_dragging.n_actions = 2;
	menu_dragging.actions = 
		(Action*) g_malloc (sizeof (Action) * menu_dragging.n_actions);

	menu_dragging.actions [0] = CANCEL;
	menu_dragging.actions [1] = DROP;
}
