/* 
 * enPathia
 * Eneso Tecnología de Adaptación, S.L. 
 */

#include <glib.h>
#include <glib/gi18n.h>

#ifdef G_OS_WIN32
#include <windows.h>
#endif

#include <cairo.h>
#include <gtk/gtk.h>
#include <string.h>
#include "edecor.h"
#include "calibration.h"
#include "config.h"
#include "enpathia.h"
#include "help.h"
#include "keyboard.h"
#include "misc.h"
#include "menu.h"
#include "profile.h"
#include "users.h"
#include "window.h"

gboolean	disable_main_window	= FALSE;


static struct
{
	GtkWidget* 	window_main;
	GtkWidget* 	vbox_main;
	GtkWidget* 	menubar;
	GtkWidget*	menuitem_config [5];
	GtkWidget*	radiomenuitem_lang [N_LANGUAGES];
	GtkWidget*	radiomenuitem_keyb_layout [N_KEYB_LAYOUTS];
	GtkWidget* 	button_profile;
	GtkWidget* 	notebook_main;
	GtkWidget*	button_new_shortcut;
	GtkWidget* 	button_shortcut [MAX_SHORTCUTS];
	GtkWidget*	menu_button_shortcut;
	GtkWidget*	notebook_config;
	GtkWidget*	togglebutton_config_menu [5];
	GtkWidget*	button_calibrate;
	GtkWidget*	radiobutton_mov_speed [3];
	GtkWidget*	checkbutton_auto_calibrate;
	GtkWidget*	radiobutton_sensib [4][5];
	GtkWidget*	radiobutton_activation_time;
	GtkWidget*	radiobutton_activation_switch;
	GtkWidget*	radiobutton_menu_time [4];
	GtkWidget*	radiobutton_keyb_time [4];
	GtkWidget*	checkbutton_can_disable_menu;
	GtkWidget*	checkbutton_auto_click;
	GtkWidget*	radiobutton_long_press_time [4];
	GtkWidget*	frame_activation_time;
	GtkWidget*	label_activation_time;
	GtkWidget*	label_menu_time;
	GtkWidget*	label_keyb_time;
	GtkWidget*	label_can_disable_menu;
	GtkWidget*	frame_activation_switch;
	GtkWidget*	label_activation_switch;
	GtkWidget*	label_long_press_time;
	GtkWidget*	radiobutton_menu_size [3];
	GtkWidget*	checkbutton_menu_use_images;
	GtkWidget*	radiobutton_keyb_size [3];
	GtkWidget*	checkbutton_keyb_show_dictionary;
	GtkWidget*	checkbutton_keyb_show_layout_bar;
	GtkWidget*	alignment_combobox_switch [4];
	GtkWidget*	combobox_switch [4];
	GtkWidget*	vbox_help;

	GtkWidget* 	window_about;
	GtkWidget* 	vbox_about;
	GtkWidget* 	label_version;
	GtkWidget* 	image_enpathia;
	GtkWidget* 	image_eneso;

	GtkWidget* 	dialog_shortcut;
	GtkWidget* 	vbox_shortcut;
	GtkWidget* 	entry_title;
	GtkWidget* 	entry_uri;
}
widgets;

static guint	current_config_section = 0;
static guint	active_shortcut_button;

static gboolean loading = FALSE;

static void
button_help_clicked_cb 	(GtkWidget *widget, gpointer data);

static void
update_configuration_page			(void);
static void
show_configuration_section			(guint section);
static void
enable_config_activation_widgets	(void);

static void
update_config_menu_items	(guint current_page);
static void
create_shortcut_button		(void);
static gboolean
edit_shortcut_button		(void);
static void
remove_shortcut_button		(void);

static gboolean
confirm_quit	(void);

G_MODULE_EXPORT void
config_change_cb	(GtkWidget *widget, gpointer data);

#ifdef __MACH__
static gboolean
osx_block_termination	(GtkOSXApplication* app, gpointer user_data);
#endif	/* __MACH__ */


	void
load_main_window (gint width, gint height)
{
	GtkBuilder*	builder;
	GtkWidget*	widget;
	gchar*		data_file;
	gint 		i, j;
	gchar		text [100];

	/* Load the interface files */
	builder = gtk_builder_new ();

	data_file = get_data_file ("interface/main.glade");
	gtk_builder_add_from_file (builder, data_file, NULL);
	g_free (data_file);

	data_file = get_data_file ("interface/about.glade");
	gtk_builder_add_from_file (builder, data_file, NULL);
	g_free (data_file);

	data_file = get_data_file ("interface/shortcut.glade");
	gtk_builder_add_from_file (builder, data_file, NULL);
	g_free (data_file);

	/* Get the widgets */
#define GET_WIDGET(widget,name) \
	widgets.widget = GTK_WIDGET (gtk_builder_get_object (builder, name)); \
	if (widgets.widget == NULL) g_warning ("Widget not found: %s", name);

	GET_WIDGET (window_main, "window_main");
	GET_WIDGET (menubar, "menubar");
	GET_WIDGET (vbox_main, "vbox_main");
	GET_WIDGET (vbox_help, "vbox_help");
	GET_WIDGET (button_profile, "button_profile");
	GET_WIDGET (button_new_shortcut, "button_new_shortcut");
	GET_WIDGET (menu_button_shortcut, "menu_button_shortcut");
	GET_WIDGET (notebook_main, "notebook_main");
	GET_WIDGET (notebook_config, "notebook_config");
	GET_WIDGET (button_calibrate, "button_calibrate");
	GET_WIDGET (frame_activation_time, "frame_activation_time");
	GET_WIDGET (label_activation_time, "label_activation_time");
	GET_WIDGET (checkbutton_auto_calibrate, "checkbutton_auto_calibrate");
	GET_WIDGET (checkbutton_can_disable_menu, "checkbutton_can_disable_menu");
	GET_WIDGET (checkbutton_auto_click, "checkbutton_auto_click");
	GET_WIDGET (label_menu_time, "label_menu_time");
	GET_WIDGET (label_keyb_time, "label_keyb_time");
	GET_WIDGET (label_can_disable_menu, "label_can_disable_menu");
	GET_WIDGET (frame_activation_switch, "frame_activation_switch");
	GET_WIDGET (label_activation_switch, "label_activation_switch");
	GET_WIDGET (label_long_press_time, "label_long_press_time");
	GET_WIDGET (checkbutton_menu_use_images, "checkbutton_menu_use_images");
	GET_WIDGET (checkbutton_keyb_show_dictionary, "checkbutton_keyb_show_dictionary");
	GET_WIDGET (checkbutton_keyb_show_layout_bar, "checkbutton_keyb_show_layout_bar");
	GET_WIDGET (radiobutton_activation_time, "radiobutton_activation_time");
	GET_WIDGET (radiobutton_activation_switch, "radiobutton_activation_switch");

	for (i = 0; i < MAX_SHORTCUTS; i++)
	{
		sprintf (text, "button_shortcut_%d", i + 1);
		GET_WIDGET (button_shortcut [i], text);
	}

	for (i = 0; i < 3; i++)
	{
		sprintf (text, "radiobutton_mov_speed_%d", i + 1);
		GET_WIDGET (radiobutton_mov_speed [i], text);
		sprintf (text, "radiobutton_menu_size_%d", i + 1);
		GET_WIDGET (radiobutton_menu_size [i], text);
		sprintf (text, "radiobutton_keyb_size_%d", i + 1);
		GET_WIDGET (radiobutton_keyb_size [i], text);
	}

	for (i = 0; i < 4; i++)
	{
		sprintf (text, "alignment_combobox_switch_%d", i + 1);
		GET_WIDGET (alignment_combobox_switch [i], text);
		sprintf (text, "radiobutton_menu_time_%d", i + 1);
		GET_WIDGET (radiobutton_menu_time [i], text);
		sprintf (text, "radiobutton_keyb_time_%d", i + 1);
		GET_WIDGET (radiobutton_keyb_time [i], text);
		sprintf (text, "radiobutton_long_press_time_%d", i + 1);
		GET_WIDGET (radiobutton_long_press_time [i], text);

		for (j = 0; j < 5; j++)
		{
			sprintf (text, "radiobutton_sensib_%d_%d", i + 1, j + 1);
			GET_WIDGET (radiobutton_sensib [i][j], text);
		}
	}

	for (i = 0; i < 5; i++)
	{
		sprintf (text, "menuitem_config_%d", i + 1);
		GET_WIDGET (menuitem_config [i], text);
		sprintf (text, "togglebutton_config_menu_%d", i + 1);
		GET_WIDGET (togglebutton_config_menu [i], text);
	}

	for (i = 0; i < N_LANGUAGES; i++)
	{
		sprintf (text, "radiomenuitem_lang_%s", language_name [i]);
		GET_WIDGET (radiomenuitem_lang [i], text);

		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (widgets.radiomenuitem_lang [i]),
				i == current_language);
	}

	for (i = 0; i < N_KEYB_LAYOUTS; i++)
	{
		sprintf (text, "radiomenuitem_keyb_layout_%s", keyb_layout_name [i]);
		GET_WIDGET (radiomenuitem_keyb_layout [i], text);

		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (widgets.radiomenuitem_keyb_layout [i]),
				i == current_keyb_layout);
	}

	GET_WIDGET (window_about, "window_about");
	GET_WIDGET (vbox_about, "vbox_about");
	GET_WIDGET (label_version, "label_version");
	GET_WIDGET (image_enpathia, "image_enpathia");
	GET_WIDGET (image_eneso, "image_eneso");

	GET_WIDGET (dialog_shortcut, "dialog_shortcut");
	GET_WIDGET (vbox_shortcut, "vbox_shortcut");
	GET_WIDGET (entry_title, "entry_title");
	GET_WIDGET (entry_uri, "entry_uri");

#undef GET_WIDGET

	gtk_builder_connect_signals (builder, &widgets);

	/* Destroying the builder breaks menu_button_shortcut */
	/* g_object_unref (G_OBJECT (builder)); */

	gtk_window_set_default_size (GTK_WINDOW (widgets.window_main), width, height);	

	data_file = get_data_file ("/images/enpathia.png");
	gtk_image_set_from_file (GTK_IMAGE (widgets.image_enpathia), data_file);
	g_free (data_file);

	data_file = get_data_file ("/images/eneso.png");
	gtk_image_set_from_file (GTK_IMAGE (widgets.image_eneso), data_file);
	g_free (data_file);

	sprintf (text, "<span font=\"" FONT " 13px\" foreground=\"#404040\">%s %s</span>", _("Version"), VERSION);
	gtk_label_set_markup (GTK_LABEL (widgets.label_version), text);

	/* Switch configuration comboboxes */
	for (i = 0; i < 4; i++)
	{
		widgets.combobox_switch [i] = gtk_combo_box_new_text ();

		g_signal_connect (widgets.combobox_switch [i],
				"changed", G_CALLBACK (config_change_cb), NULL);

		gtk_combo_box_append_text (GTK_COMBO_BOX (widgets.combobox_switch [i]), 
				_("Show the menu"));
		for (j = 2; j < 5; j++)
		{
			gtk_combo_box_append_text (GTK_COMBO_BOX (widgets.combobox_switch [i]), 
					action_text [j]);
		}

		gtk_container_add (GTK_CONTAINER (widgets.alignment_combobox_switch [i]),
				widgets.combobox_switch [i]);
		gtk_widget_show (widgets.combobox_switch [i]);
	}

	/* Help page */
	for (i = 0; i < N_HELP_SHORTCUTS; ++i)
	{
		widget = gtk_button_new_with_label (gettext (help_shortcut_title [i]));
		gtk_widget_set_size_request (widget, -1, 30);
		gtk_box_pack_start (GTK_BOX (widgets.vbox_help), widget, FALSE, FALSE, 5);	

		g_signal_connect (widget, "clicked",
				G_CALLBACK (button_help_clicked_cb), (void*) i);

		gtk_widget_show (widget);
	}

	load_configuration_values ();

	edecor_decorate_window (widgets.window_main, widgets.vbox_main);
	edecor_decorate_window (widgets.window_about, widgets.vbox_about);
	edecor_decorate_window (widgets.dialog_shortcut, widgets.vbox_shortcut);

#ifdef __MACH__
	/* Hide the window menubar and put it on the desktop bar */
	gtk_osxapplication_set_menu_bar (gtk_osx_app, GTK_MENU_SHELL (widgets.menubar));
	gtk_widget_hide (widgets.menubar);

	/* Make sure the window is not destroyed when closed, only hidden */
	g_signal_connect (widgets.window_main, 
			"delete-event", G_CALLBACK (gtk_widget_hide_on_delete), NULL);

	/* Connect to the 'Quit' signal on the default dock menu */
	g_signal_connect (gtk_osx_app, 
			"NSApplicationBlockTermination", G_CALLBACK (osx_block_termination), NULL);
#endif	/* __MACH__ */
}

	void
show_main_window (gint x, gint y)
{
	gint	width, height;

	if (disable_main_window)
		return;

	gtk_window_get_size (GTK_WINDOW (widgets.window_main), &width, &height);

	if (x == G_MAXINT)
		x = (screen_width - width) >> 1;
	if (y == G_MAXINT)
		y = (screen_height - height) >> 1;

	if (x < 0)
		x = 0;
	else if (x > screen_width - width)
		x = screen_width - width;

	if (y < 0)
		y = 0;
	else if (y > screen_height - height)
		y = screen_height - height;

	gtk_window_move (GTK_WINDOW (widgets.window_main), x, y);
	gtk_window_present (GTK_WINDOW (widgets.window_main));
}

	void
update_main_window (void)
{
	gint	i;

	/* Current profile name */
	gtk_button_set_label (GTK_BUTTON (widgets.button_profile), current_profile_name);

	/* Window title */
	gtk_window_set_title (GTK_WINDOW (widgets.window_main),
			device_connected ? _("enPathia") : _("enPathia (disconnected)"));

	/* Shortcuts */
	gtk_widget_set_sensitive (widgets.button_new_shortcut, 
			current_profile.main_window.n_shortcuts < MAX_SHORTCUTS);

	for (i = 0; i < MAX_SHORTCUTS; i++)
	{
		if (i <	current_profile.main_window.n_shortcuts)
		{
			gtk_button_set_label (GTK_BUTTON (widgets.button_shortcut [i]),
					strdup (current_profile.main_window.shortcut_title [i]));
			gtk_widget_show (widgets.button_shortcut [i]);
		}
		else
			gtk_widget_hide (widgets.button_shortcut [i]);
	}

	/* Configuration */
	update_configuration_page ();
	update_config_menu_items (
			gtk_notebook_get_current_page (GTK_NOTEBOOK (widgets.notebook_main)));

	gtk_widget_queue_draw (widgets.window_main);
}

	void
quit_with_confirmation (void)
{
	if (confirm_quit ())
		gtk_main_quit ();
}

	void
load_configuration_values (void)
{
	gint	i, j, n;

	/* 
	 * Load the current profile values
	 */

	loading = TRUE;

	/* Movement */
	for (i = 0; i < 3; i++)
	{
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_mov_speed [i]),
				current_profile.config.movement_speed == i);
	}

	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.checkbutton_auto_calibrate),
			current_profile.config.auto_calibrate == TRUE);

	/* Sensitivity */
	for (i = 0; i < 4; i++)
	{
		for (j = 0; j < 5; j++)
		{
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_sensib [i][j]),
					current_profile.config.sensitivity [i] == j);
		}
	}

	/* Activation */
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_activation_time),
			current_profile.config.use_switches == FALSE);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_activation_switch),
			current_profile.config.use_switches == TRUE);

	for (i = 0; i < 4; i++)
	{
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_menu_time [i]),
				current_profile.config.menu_time == i);

		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_keyb_time [i]),
				current_profile.config.keyb_time == i);

		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_long_press_time [i]),
				current_profile.config.long_press_time == i);
	}

	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.checkbutton_can_disable_menu),
			current_profile.config.can_disable_menu == TRUE);

	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.checkbutton_auto_click),
			current_profile.config.auto_click == TRUE);

	enable_config_activation_widgets ();

	/* Size */
	for (i = 0; i < 3; i++)
	{
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_menu_size [i]),
				current_profile.config.menu_size == i);
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_keyb_size [i]),
				current_profile.config.keyb_size == i);
	}

	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.checkbutton_menu_use_images),
			current_profile.config.menu_use_images);

	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.checkbutton_keyb_show_dictionary),
			current_profile.config.keyb_show_dictionary);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.checkbutton_keyb_show_layout_bar),
			current_profile.config.keyb_show_layout_bar);

	/* Switches */
	for (i = 0; i < 4; i++)
	{
		n = current_profile.config.switch_action [i];
		if (n > 1)
			n--;

		gtk_combo_box_set_active (GTK_COMBO_BOX (widgets.combobox_switch [i]), n);

	}

	loading = FALSE;
}

	void
show_main_window_config (void)
{
	gtk_notebook_set_current_page (GTK_NOTEBOOK (widgets.notebook_main), 1);
	show_configuration_section (0);

	gtk_window_present (GTK_WINDOW (widgets.window_main));
}

	G_MODULE_EXPORT gboolean
window_main_configure_event_cb (GtkWidget *widget, GdkEventConfigure *event, gpointer data)
{
	current_profile.main_window.x = event->x;
	current_profile.main_window.y = event->y;
	current_profile.main_window.width = event->width;
	current_profile.main_window.height = event->height;

	return FALSE;
}

	G_MODULE_EXPORT gboolean
window_main_show_cb (GtkWidget *widget, gpointer data)
{
	/* When the window open or closes remember it in the profile */
	current_profile.main_window.show = TRUE;

	return FALSE;
}

	G_MODULE_EXPORT gboolean
window_main_hide_cb (GtkWidget *widget, gpointer data)
{
	current_profile.main_window.show = FALSE;

	return FALSE;
}

	G_MODULE_EXPORT void
notebook_switch_page_cb (GtkNotebook* notebook, GtkNotebookPage* page, guint num, gpointer user_data)
{
	update_config_menu_items (num);
}

	G_MODULE_EXPORT void
menuitem_edit_profiles_activate_cb (GtkMenuItem* menuitem, gpointer data)
{
	show_users_dialog ();
}

	G_MODULE_EXPORT void
menuitem_import_profile_activate_cb (GtkMenuItem* menuitem, gpointer data)
{
	import_profile_dialog ();
}

	G_MODULE_EXPORT void
menuitem_export_profile_activate_cb (GtkMenuItem* menuitem, gpointer data)
{
	export_profile_dialog (current_profile_name);
}

	G_MODULE_EXPORT void
menuitem_quit_activate_cb (GtkMenuItem* menuitem, gpointer data)
{
	quit_with_confirmation ();
}

	G_MODULE_EXPORT void
menuitem_new_shortcut_activate_cb (GtkMenuItem* menuitem, gpointer data)
{
	create_shortcut_button ();
}

	G_MODULE_EXPORT void
menuitem_remove_all_shortcuts_activate_cb (GtkMenuItem* menuitem, gpointer data)
{
	if (dialog_yes_no (_("Shortcuts"), _("Are you sure you want to remove all the shortcuts?")))
	{
		while (current_profile.main_window.n_shortcuts > 0)
		{
			active_shortcut_button = 0;
			remove_shortcut_button ();
		}

		update_main_window ();
	}
}

	static void
radiomenuitem_lang_activate_cb (GtkMenuItem* menuitem, gint language)
{
	if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (menuitem)))
	{
		if (language != current_language)
		{
			set_env_var ("LANG", language_name [language]);
			current_language = language;

			update_config_file ();

			dialog_info (_("Language"),
					_("You need to restart the application to apply this change"));
		}
	}
}

#define RADIOMENUITEM_LANG_ACTIVATE_CB(language,cb) \
	G_MODULE_EXPORT	void cb (GtkMenuItem* menuitem, gpointer data) \
{ radiomenuitem_lang_activate_cb(menuitem, language); }

RADIOMENUITEM_LANG_ACTIVATE_CB (LANG_EN, radiomenuitem_lang_en_activate_cb);
RADIOMENUITEM_LANG_ACTIVATE_CB (LANG_EL, radiomenuitem_lang_el_activate_cb);
RADIOMENUITEM_LANG_ACTIVATE_CB (LANG_ES, radiomenuitem_lang_es_activate_cb);
RADIOMENUITEM_LANG_ACTIVATE_CB (LANG_DE, radiomenuitem_lang_de_activate_cb);
RADIOMENUITEM_LANG_ACTIVATE_CB (LANG_HE, radiomenuitem_lang_he_activate_cb);
RADIOMENUITEM_LANG_ACTIVATE_CB (LANG_IT, radiomenuitem_lang_it_activate_cb);
RADIOMENUITEM_LANG_ACTIVATE_CB (LANG_PT, radiomenuitem_lang_pt_activate_cb);
RADIOMENUITEM_LANG_ACTIVATE_CB (LANG_FR, radiomenuitem_lang_fr_activate_cb);

#undef MENUITEM_LANG_ACTIVATE_CB

	static void
radiomenuitem_keyb_layout_activate_cb (GtkMenuItem* menuitem, gint keyb_layout)
{
	if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (menuitem)))
	{
		if (keyb_layout != current_keyb_layout)
		{
			current_keyb_layout = keyb_layout;
			reset_keyboard ();

			update_config_file ();
		}
	}
}

#define RADIOMENUITEM_KEYB_LAYOUT_ACTIVATE_CB(keyb_layout,cb) \
	G_MODULE_EXPORT	void cb (GtkMenuItem* menuitem, gpointer data) \
{ radiomenuitem_keyb_layout_activate_cb(menuitem, keyb_layout); }

RADIOMENUITEM_KEYB_LAYOUT_ACTIVATE_CB (KEYB_LAYOUT_EN_US, radiomenuitem_keyb_layout_en_US_activate_cb);
RADIOMENUITEM_KEYB_LAYOUT_ACTIVATE_CB (KEYB_LAYOUT_EN_UK, radiomenuitem_keyb_layout_en_UK_activate_cb);
RADIOMENUITEM_KEYB_LAYOUT_ACTIVATE_CB (KEYB_LAYOUT_EL, radiomenuitem_keyb_layout_el_activate_cb);
RADIOMENUITEM_KEYB_LAYOUT_ACTIVATE_CB (KEYB_LAYOUT_ES, radiomenuitem_keyb_layout_es_activate_cb);
RADIOMENUITEM_KEYB_LAYOUT_ACTIVATE_CB (KEYB_LAYOUT_DE, radiomenuitem_keyb_layout_de_activate_cb);
RADIOMENUITEM_KEYB_LAYOUT_ACTIVATE_CB (KEYB_LAYOUT_HE, radiomenuitem_keyb_layout_he_activate_cb);
RADIOMENUITEM_KEYB_LAYOUT_ACTIVATE_CB (KEYB_LAYOUT_IT, radiomenuitem_keyb_layout_it_activate_cb);
RADIOMENUITEM_KEYB_LAYOUT_ACTIVATE_CB (KEYB_LAYOUT_PT, radiomenuitem_keyb_layout_pt_activate_cb);
RADIOMENUITEM_KEYB_LAYOUT_ACTIVATE_CB (KEYB_LAYOUT_FR, radiomenuitem_keyb_layout_fr_activate_cb);

#undef MENUITEM_KEYB_LAYOUT_ACTIVATE_CB

	static void
menuitem_config_activate_cb (GtkMenuItem* menuitem, gint section)
{
	if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (menuitem)))
	{
		gtk_notebook_set_current_page (GTK_NOTEBOOK (widgets.notebook_main), 1);			
		show_configuration_section (section);
	}
}

#define MENUITEM_CONFIG_ACTIVATE_CB(section,cb) \
	G_MODULE_EXPORT	void cb (GtkMenuItem* menuitem, gpointer data) \
{ menuitem_config_activate_cb(menuitem, section); }

MENUITEM_CONFIG_ACTIVATE_CB (0, menuitem_config_1_activate_cb);
MENUITEM_CONFIG_ACTIVATE_CB (1, menuitem_config_2_activate_cb);
MENUITEM_CONFIG_ACTIVATE_CB (2, menuitem_config_3_activate_cb);
MENUITEM_CONFIG_ACTIVATE_CB (3, menuitem_config_4_activate_cb);
MENUITEM_CONFIG_ACTIVATE_CB (4, menuitem_config_5_activate_cb);

#undef MENUITEM_CONFIG_ACTIVATE_CB

	G_MODULE_EXPORT void
menuitem_quick_guide_activate_cb (GtkMenuItem* menuitem, gpointer data)
{
	show_help (0);
}

	G_MODULE_EXPORT void
menuitem_about_activate_cb (GtkMenuItem* menuitem, gpointer data)
{
	gtk_window_present (GTK_WINDOW (widgets.window_about));
}

	G_MODULE_EXPORT void
button_profile_clicked_cb (GtkWidget* widget, gpointer data)
{
	show_users_dialog ();
}

	G_MODULE_EXPORT void
button_new_shortcut_clicked_cb (GtkWidget *widget, gpointer data)
{
	create_shortcut_button ();
}

	static void
button_shortcut_button_release_cb (gint button, GdkEventButton* event)
{
	if (event->button == 1)
	{
		/* Launch the URI associated to the shortcut */
		if (strcmp (current_profile.main_window.shortcut_uri [button], "") != 0)
			execute_uri (current_profile.main_window.shortcut_uri [button]);
	}
	else
	{
		active_shortcut_button = button;

		gtk_menu_popup (GTK_MENU (widgets.menu_button_shortcut), NULL, NULL, NULL, NULL,
				event->button, event->time);
	}
}

#define BUTTON_SHORTCUT_BUTTON_RELEASE_CB(button,cb) \
	G_MODULE_EXPORT	void cb (GtkWidget* widget, GdkEventButton* event, gpointer data) \
{ button_shortcut_button_release_cb(button, event); }

BUTTON_SHORTCUT_BUTTON_RELEASE_CB (0, button_shortcut_1_button_release_event_cb);
BUTTON_SHORTCUT_BUTTON_RELEASE_CB (1, button_shortcut_2_button_release_event_cb);
BUTTON_SHORTCUT_BUTTON_RELEASE_CB (2, button_shortcut_3_button_release_event_cb);
BUTTON_SHORTCUT_BUTTON_RELEASE_CB (3, button_shortcut_4_button_release_event_cb);
BUTTON_SHORTCUT_BUTTON_RELEASE_CB (4, button_shortcut_5_button_release_event_cb);
BUTTON_SHORTCUT_BUTTON_RELEASE_CB (5, button_shortcut_6_button_release_event_cb);
BUTTON_SHORTCUT_BUTTON_RELEASE_CB (6, button_shortcut_7_button_release_event_cb);
BUTTON_SHORTCUT_BUTTON_RELEASE_CB (7, button_shortcut_8_button_release_event_cb);
BUTTON_SHORTCUT_BUTTON_RELEASE_CB (8, button_shortcut_9_button_release_event_cb);
BUTTON_SHORTCUT_BUTTON_RELEASE_CB (9, button_shortcut_10_button_release_event_cb);
BUTTON_SHORTCUT_BUTTON_RELEASE_CB (10, button_shortcut_11_button_release_event_cb);
BUTTON_SHORTCUT_BUTTON_RELEASE_CB (11, button_shortcut_12_button_release_event_cb);

#undef BUTTON_SHORTCUT_BUTTON_RELEASE_CB

	G_MODULE_EXPORT void
menuitem_shortcut_edit_activate_cb (GtkMenuItem* menuitem, gpointer data)
{
	edit_shortcut_button ();

	update_main_window ();
}

	G_MODULE_EXPORT void
menuitem_shortcut_remove_activate_cb (GtkMenuItem* menuitem, gpointer data)
{
	remove_shortcut_button ();

	update_main_window ();
}

	static void
button_help_clicked_cb (GtkWidget *widget, gpointer data)
{
	show_help ((int) data);
}

	void
update_config_menu_items (guint current_page)
{
	gint	i;

	for (i = 0; i < 5; i++)
	{
		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (widgets.menuitem_config [i]),
				current_page == 1 && current_config_section == i);
	}

	gtk_check_menu_item_set_active (
			GTK_CHECK_MENU_ITEM (widgets.radiomenuitem_keyb_layout [current_keyb_layout]), TRUE);

	gtk_widget_set_sensitive (widgets.menuitem_config [4], current_profile.config.use_switches);
}

	static void
create_shortcut_button (void)
{
	gint	n;

	n = current_profile.main_window.n_shortcuts;
	current_profile.main_window.n_shortcuts++;

	strcpy (current_profile.main_window.shortcut_title [n], "");
	strcpy (current_profile.main_window.shortcut_uri [n], "");

	update_main_window ();

	active_shortcut_button = n;
	edit_shortcut_button ();
}

	static gboolean
edit_shortcut_button (void)
{
	gtk_entry_set_text (GTK_ENTRY (widgets.entry_title),
			current_profile.main_window.shortcut_title [active_shortcut_button]);
	gtk_entry_set_text (GTK_ENTRY (widgets.entry_uri),
			current_profile.main_window.shortcut_uri [active_shortcut_button]);

	gtk_widget_show (widgets.dialog_shortcut);

	return TRUE;
}

	G_MODULE_EXPORT void
dialog_shortcut_response_cb (GtkDialog *dialog, gint response_id, gpointer data)
{
	gtk_widget_hide (GTK_WIDGET (dialog));

	if (response_id == 1)
	{
		strcpy (current_profile.main_window.shortcut_title [active_shortcut_button],
				gtk_entry_get_text (GTK_ENTRY (widgets.entry_title)));
		strcpy (current_profile.main_window.shortcut_uri [active_shortcut_button],
				gtk_entry_get_text (GTK_ENTRY (widgets.entry_uri)));

		update_main_window ();
	}

	active_shortcut_button = -1;
}

	G_MODULE_EXPORT void
button_browse_clicked_cb (GtkWidget *widget, gpointer data)
{
	gchar*	file;
	gchar	uri [256];

	if (file_chooser (&file, _("Select file"), TRUE, NULL,
				_("All files (*.*)"), "*.*"))
	{
		if (file)
		{
			strcpy (uri, "file://");
			strcat (uri, file);

			gtk_entry_set_text (GTK_ENTRY (widgets.entry_uri), uri);
		}

		g_free (file);
	}
}

	static void
remove_shortcut_button (void)
{
	gint	i;

	for (i = active_shortcut_button; i < MAX_SHORTCUTS - 1; i++)
	{
		strcpy (current_profile.main_window.shortcut_title [i],
				current_profile.main_window.shortcut_title [i + 1]);
		strcpy (current_profile.main_window.shortcut_uri [i],
				current_profile.main_window.shortcut_uri [i + 1]);
	}

	current_profile.main_window.n_shortcuts--;

	active_shortcut_button = -1;
}

	static void
update_configuration_page (void)
{
	gtk_widget_set_sensitive (widgets.button_calibrate, device_connected);
}

	static void
show_configuration_section (guint section)
{
	gint	i;

	if (widgets.notebook_config == NULL)
		return;

	if (gtk_notebook_get_current_page (GTK_NOTEBOOK (widgets.notebook_config)) != section)
	{
		gtk_notebook_set_current_page (GTK_NOTEBOOK (widgets.notebook_config), section);

		for (i = 0; i < 5; i++)
		{
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.togglebutton_config_menu [i]),
					i == section);
		}

		current_config_section = section;
		update_main_window ();
	}
}

	static gboolean
confirm_quit (void)
{
	return dialog_yes_no (_("Are you sure you want to quit?"),
			_("Doing so will stop from controlling your computer with enPathia."));
}

#ifdef __MACH__
	static gboolean
osx_block_termination (GtkOSXApplication* app, gpointer user_data)
{
	return !confirm_quit ();
}
#endif	/* __MACH__ */

	G_MODULE_EXPORT void
config_change_cb (GtkWidget *widget, gpointer data)
{
	Profile	profile;
	guint	i, j;
	gint	n;

	if (loading)
		return;

	/* 
	 * Fill a profile up with current values
	 */

	/* Movemente */
	for (i = 0; i < 3; i++)
	{
		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_mov_speed [i])))
			profile.config.movement_speed = i;
	}

	profile.config.auto_calibrate =
		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.checkbutton_auto_calibrate));

	/* Sensitivity */
	for (i = 0; i < 4; i++)
	{
		for (j = 0; j < 5; j++)
		{
			if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_sensib [i][j])))
			{
				profile.config.sensitivity [i] = j;
				break;
			}
		}
	}

	/* Activation */
	profile.config.use_switches =
		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_activation_switch));

	for (i = 0; i < 4; i++)
	{
		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_menu_time [i])))
			profile.config.menu_time = i;

		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_keyb_time [i])))
			profile.config.keyb_time = i;

		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_long_press_time [i])))
			profile.config.long_press_time = i;
	}

	profile.config.can_disable_menu =
		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.checkbutton_can_disable_menu));

	profile.config.auto_click =
		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.checkbutton_auto_click));

	/* Size */
	for (i = 0; i < 3; i++)
	{
		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_menu_size [i])))
			profile.config.menu_size = i;
		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_keyb_size [i])))
			profile.config.keyb_size = i;
	}

	profile.config.menu_use_images =
		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.checkbutton_menu_use_images));

	profile.config.keyb_show_dictionary =
		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.checkbutton_keyb_show_dictionary));
	profile.config.keyb_show_layout_bar =
		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.checkbutton_keyb_show_layout_bar));

	/* Switches */
	for (i = 0; i < 4; i++)
	{
		n = gtk_combo_box_get_active (GTK_COMBO_BOX (widgets.combobox_switch [i]));
		if (n > 0)
			n++;

		profile.config.switch_action [i] = n;

	}

	enable_config_activation_widgets ();

	/* If it is different from the current one save it and send the configuration to 
	 * the device */
	if (memcmp (&profile.config, &current_profile.config, sizeof (profile.config)) != 0)
	{
		current_profile.config = profile.config;

		update_main_window ();

		set_menu_size (current_profile.config.menu_size);
		set_keyboard_size (current_profile.config.keyb_size);

		send_cfg_speed ();
		send_cfg_time ();
		send_cfg_switches ();
		send_cfg_sensitivity ();
		send_cfg_calibration ();
	}
}

	G_MODULE_EXPORT void
button_calibrate_clicked_cb (GtkWidget *widget, gpointer data)
{
	calibrate ();
}

	static void
togglebutton_config_menu_clicked_cb (GtkWidget *widget, gint n)
{
	/* Open the notebook page associated to the pressed menu button.
	 * Simulate the behaviour of radiobuttons. */
	if (gtk_notebook_get_current_page (GTK_NOTEBOOK (widgets.notebook_config)) != n)
	{
		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
			show_configuration_section (n);
	}
	else
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
}

#define TOGGLEBUTTON_CONFIG_MENU_CLICKED_CB(n,cb) \
	G_MODULE_EXPORT	void cb (GtkWidget* widget, gpointer data) \
{ togglebutton_config_menu_clicked_cb(widget, n); }

TOGGLEBUTTON_CONFIG_MENU_CLICKED_CB (0, togglebutton_config_menu_1_clicked_cb);
TOGGLEBUTTON_CONFIG_MENU_CLICKED_CB (1, togglebutton_config_menu_2_clicked_cb);
TOGGLEBUTTON_CONFIG_MENU_CLICKED_CB (2, togglebutton_config_menu_3_clicked_cb);
TOGGLEBUTTON_CONFIG_MENU_CLICKED_CB (3, togglebutton_config_menu_4_clicked_cb);
TOGGLEBUTTON_CONFIG_MENU_CLICKED_CB (4, togglebutton_config_menu_5_clicked_cb);

#undef TOGGLEBUTTON_CONFIG_MENU_CLICKED_CB

	static void
enable_config_activation_widgets (void)
{
	gint	i;

	gboolean activ_time = 
		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.radiobutton_activation_time));

	gtk_widget_set_sensitive (widgets.frame_activation_time, activ_time);
	gtk_widget_set_sensitive (widgets.label_activation_time, activ_time);

	gtk_widget_set_sensitive (widgets.label_menu_time, activ_time);
	gtk_widget_set_sensitive (widgets.label_keyb_time, activ_time);

	gtk_widget_set_sensitive (widgets.frame_activation_switch, !activ_time);
	gtk_widget_set_sensitive (widgets.label_activation_switch, !activ_time);

	gtk_widget_set_sensitive (widgets.label_long_press_time, !activ_time);

	for (i = 0; i < 4; i++)
	{
		gtk_widget_set_sensitive (widgets.radiobutton_menu_time[i], activ_time);
		gtk_widget_set_sensitive (widgets.radiobutton_keyb_time [i], activ_time);
		gtk_widget_set_sensitive (widgets.radiobutton_long_press_time [i], !activ_time);
	}

	gtk_widget_set_sensitive (widgets.checkbutton_can_disable_menu, activ_time);
	gtk_widget_set_sensitive (widgets.label_can_disable_menu, activ_time &&
			gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.checkbutton_can_disable_menu)));

	gtk_widget_set_sensitive (widgets.checkbutton_auto_click, activ_time);

	gtk_widget_set_sensitive (widgets.togglebutton_config_menu [4], !activ_time);

	update_main_window ();
}
